#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2024 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import ast
import base64
import hashlib
import json
import multiprocessing as mp
import os
import pandas as pd
import psycopg2
import re
import shutil

from datetime import datetime
from flask import Flask, request, Response
from netaddr import IPAddress
from openapi_spec_validator import validate_spec
from psycopg2.extras import RealDictCursor
from time import time, time_ns
from urllib.parse import unquote
from zipfile import ZipFile

from core import cc_upd_db
from core import close_db_connection
from core import get_agent_ip
from core import get_attack_uri, get_attack_stats_uri
from core import get_auth, get_waf_id
from core import get_data_from_json
from core import init_db
from core import init_db_connection
from core import log
from core import memc
from core import memc_lock
from core import ro_mode
from core import set_attack_processing
from core import set_attack_uri
from core import set_scanner_uri
from core import vhost_normalization, idna_vhost_list_normalization

from ws import dyn_bl_cols, dyn_bl_value_validation
from ws import dyn_erl_cols, dyn_erl_value_validation, dyn_erl_type_data, dyn_erl_type_boolean
from ws import dyn_sync_ban_memc_ttl, dyn_sync_ban_cols, dyn_sync_ban_value_validation
from ws import dyn_wl_cols, dyn_wl_value_validation
from ws import dyn_wl_erl_cmd, dyn_settings_cols, dyn_value_validation
from ws import get_dyn_bl_uri, set_dyn_bl_uri
from ws import get_dyn_erl_hash_uri, set_dyn_wl_uri
from ws import get_dyn_erl_uri, get_dyn_bl_hash_uri, set_dyn_erl_uri
from ws import get_dyn_settings_uri, get_dyn_settings_hash_uri, set_dyn_settings_uri
from ws import get_dyn_sync_ban_uri, set_dyn_sync_ban_uri, del_dyn_sync_ban_uri
from ws import get_dyn_wl_hash_uri, get_dyn_wl_uri
from ws import get_hash, set_hash
from ws import get_mlc_settings_hash_uri, get_mlc_settings_uri, set_mlc_settings_uri
from ws import get_vhosts_list_uri, set_vhosts_list_uri
from ws import mlc_settings_cols, mlc_be_cols, mlc_value_validation
from ws import nws_target_cols, nws_target_ap_cols, nws_target_value_validation
from ws import set_nws_target_uri, del_nws_target_uri, get_nws_target_uri
from ws import settings_cmd

from ml import get_models_uri, set_models_uri, del_models_uri, rep_models_uri, get_models_list_uri
from ml import get_training_uri, set_training_uri, set_training_progress_uri, training_vhost_processing

from mt import mt_get_export, mt_get_export_md5
from mt import set_mt_data_uri, get_mt_data_uri, get_mt_hash_uri

from oa import get_openapi_exclude_paths_uri, set_openapi_exclude_paths_uri
from oa import get_openapi_schema_paths_uri, set_openapi_schema_paths_uri
from oa import get_openapi_schemas_hash_uri, get_openapi_url_hash_uri, get_openapi_exclude_paths_hash_uri
from oa import get_openapi_schemas_uri
from oa import get_openapi_url_uri, set_openapi_url_uri
from oa import oa_validation_rwo
from oa import schema_name_after_change, set_openapi_hash
from oa import set_openapi_schema_uri

from ns import get_nginx_certificate_uri, set_nginx_certificate_uri, get_nginx_certificate_hash_uri, get_nginx_certificate_secure_uri
from ns import get_nginx_dhparam_uri, set_nginx_dhparam_uri, get_nginx_dhparam_hash_uri, get_nginx_dhparam_secure_uri
from ns import get_nginx_global_uri, get_nginx_global_hash_uri, set_nginx_global_uri
from ns import get_nginx_hash, set_nginx_hash
from ns import get_nginx_status_uri
from ns import get_nginx_upstream_uri, get_nginx_upstream_hash_uri, set_nginx_upstream_uri
from ns import get_nginx_vhost_uri, get_nginx_vhost_hash_uri, set_nginx_vhost_uri
from ns import get_nginx_with_active, get_nginx_with_format, nginx_format_processing
from ns import nginx_cmd, nginx_global_cols, nginx_vhost_cols, nginx_upstream_cols, nginx_certificate_cols, nginx_dhparam_cols
from ns import nginx_type_list, nginx_type_dict, nginx_type_lod
from ns import nginx_validation

from ec import event_validation
from ec import get_event_status_uri, set_event_uri, get_event_uri, del_event_uri
from ec import set_event_cols

##
# Launch the API
##

app = Flask(__name__)
log.info('START')

# DB processing
if not ro_mode:
    init_db()


##
# Attacks result
##

@app.route(set_attack_uri, methods=['POST'])
def set_attack_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # extract client IP
        agent_ip = get_agent_ip(request)

        # processing data from the request
        try:
            data = request.get_json()
        except Exception as e:
            log.error(
                'An empty or invalid request to {} was received (client IP: {}): {}'.format(set_attack_uri, agent_ip, e)
            )
            return Response(status=400)

        # save attack in the database
        p = mp.Process(
            target=set_attack_processing,
            args=(data, agent_ip,)
        )
        p.daemon = False
        p.start()

        # response
        return Response(status=202)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_attack_uri, methods=['POST'])
def get_attack_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, _, _, _ = get_auth(key, agent_ip)

        # WAF ID check
        if not waf_id:
            return Response(status=401)

        # Update the CC
        cc_upd_db('attack')

        ##
        # Receive data from the database
        ##

        # init
        resp = {'status': 'success', 'count': 0, 'waf_id': waf_id, 'attacks': []}

        try:

            # init
            offset = 0
            limit = None
            limit_max = 3000

            # extract the limit
            try:
                k = 'limit'
                if k in jdata:
                    limit = int(jdata[k])
                    if limit > limit_max:
                        limit = limit_max
            except Exception as e:
                log.error(
                    'An error occurred while extracting the limit from {} to {}: {}'
                    .format(agent_ip, url, e)
                )
                return Response(status=400)

            # extract the offset
            if limit is not None:
                try:
                    k = 'offset'
                    if k in jdata:
                        offset = int(jdata[k]) * limit
                except Exception as e:
                    log.error(
                        'An error occurred while extracting the limit from {} to {}: {}'
                        .format(agent_ip, url, e)
                    )
                    return Response(status=400)

            # other parameters processing
            for param in ['starttime', 'endtime', 'bt', 'ip', 'vhost', 'cc']:
                try:

                    # int validation and normalization
                    if param in ['starttime', 'endtime']:
                        if param in jdata and not isinstance(jdata[param], int):
                            return Response(status=400)
                        else:
                            if param not in jdata or jdata[param] is None:
                                jdata[param] = 0 if param == 'starttime' else 9999999999

                    # list validation and normalization
                    if param in ['bt', 'ip', 'vhost', 'cc']:
                        if param in jdata and not isinstance(jdata[param], list):
                            return Response(status=400)
                        else:
                            if param not in jdata or not jdata[param]:
                                jdata[param] = '%%'
                            else:
                                # param pre-normalization
                                if param == 'cc':
                                    r = []
                                    for x in jdata[param]:
                                        x = '  ' if x is None else x
                                        r.append(x)
                                    jdata[param] = r.copy()
                                jdata[param] = '|'.join([str(x) for x in jdata[param]])

                except Exception as e:
                    log.error(
                        'An error occurred while extracting data from {} to {}: {}'
                        .format(agent_ip, url, e)
                    )
                    return Response(status=400)

            # get the count of all rows
            query = """
                SELECT
                    count(*)
                FROM
                    attack
                WHERE (
                    waf_id = %s
                        AND
                    timestamp >= to_timestamp(%s)
                        AND
                    timestamp <= to_timestamp(%s)
                        AND
                    CAST (bt AS TEXT) SIMILAR TO %s
                        AND
                    ip SIMILAR TO %s
                        AND
                    cc SIMILAR TO %s
                        AND
                    vhost SIMILAR TO %s
                )
            """
            cursor.execute(query, (
                waf_id,
                int(jdata['starttime']),
                int(jdata['endtime']),
                jdata['bt'],
                jdata['ip'],
                jdata['cc'],
                jdata['vhost'],
            ))
            res = cursor.fetchone()

            # count processing
            if res:

                # update the response data
                resp['count'] = res[0]
                resp['attacks'] = []

                # search the data
                query = """
                    SELECT
                        id, CAST(EXTRACT(EPOCH FROM timestamp) as integer), request_id, group_id, rule_id,
                        description, long_description, blocked, possible, lm, mz, bt, ip, cc, vhost, method, url, param,
                        convert_from(body, 'utf-8'), cookie, other_headers, ua, referer
                    FROM
                        attack
                    WHERE (
                        waf_id = %s
                            AND
                        timestamp >= to_timestamp(%s)
                            AND
                        timestamp <= to_timestamp(%s)
                            AND
                        CAST (bt AS TEXT) SIMILAR TO %s
                            AND
                        ip SIMILAR TO %s
                            AND
                        cc SIMILAR TO %s
                            AND
                        vhost SIMILAR TO %s
                    )
                    ORDER BY id
                    LIMIT %s
                    OFFSET %s
                """
                cursor.execute(query, (
                    waf_id,
                    int(jdata['starttime']),
                    int(jdata['endtime']),
                    jdata['bt'],
                    jdata['ip'],
                    jdata['cc'],
                    jdata['vhost'],
                    limit,
                    offset
                ))
                res = cursor.fetchall()

                # processing result
                if res:

                    # processing the result
                    for item in res:

                        # item processing
                        r = {
                            'id': item[0],
                            'timestamp': item[1],
                            'request_id': item[2],
                            'group_id': item[3],
                            'rule_id': item[4],
                            'description': item[5],
                            'long_description': item[6],
                            'blocked': item[7],
                            'possible': item[8],
                            'lm': item[9],
                            'mz': item[10],
                            'bt': item[11],
                            'ip': item[12],
                            'cc': item[13],
                            'vhost': item[14],
                            'method': item[15],
                            'url': item[16],
                            'args': item[17],
                            'body': item[18],
                            'cookie': item[19],
                            'other_headers': item[20],
                            'ua': item[21],
                            'referer': item[22]
                        }

                        # convert possible
                        if r['possible'] is None:
                            r['possible'] = False

                        # convert CC
                        if not r['cc'].strip():
                            r['cc'] = None

                        # update the response data
                        resp['attacks'].append(r)

                        # del vars
                        del r

                # del vars
                del res

        except Exception as e:
            log.error(
                'An error occurred while processing data from the database in {}: {}'
                .format(url, e)
            )

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(json.dumps(resp), status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_attack_stats_uri, methods=['POST'])
def get_attack_stats_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, _, _, _ = get_auth(key, agent_ip)

        # WAF ID check
        if not waf_id:
            return Response(status=401)

        # init
        ts = int(time())
        use_period = False
        df = pd.DataFrame()
        resp = {'status': 'success'}
        params = ['bt', 'ip', 'cc', 'vhost']
        d01_ts = ts - 86400
        d07_ts = ts - (86400 * 7)
        d30_ts = ts - (86400 * 30)

        # init param in response
        for param in params:
            resp[param] = []

        # Update the CC
        cc_upd_db('attack')

        ##
        # Receive data from the database
        ##

        try:

            # processing the parameters
            if 'startTime' in jdata and 'EndTime' in jdata:
                if isinstance(jdata['startTime'], int) and isinstance(jdata['EndTime'], int):
                    use_period = True

            # search the data
            if use_period:
                query = """
                    SELECT
                        json_agg(
                            json_build_object(
                                'timestamp', CAST(EXTRACT(EPOCH FROM timestamp) as integer),
                                'bt', bt,
                                'ip', ip,
                                'cc', cc,
                                'vhost', vhost
                            )
                        )
                    FROM
                        attack
                    WHERE (
                        waf_id = %s
                            AND
                        timestamp >= to_timestamp(%s)
                            AND
                        timestamp <= to_timestamp(%s)
                    )
                """
                cursor.execute(query, (waf_id, jdata['starttime'], jdata['endtime'],))
            else:
                # select for 1 month
                query = """
                    SELECT
                        json_agg(
                            json_build_object(
                                'timestamp', CAST(EXTRACT(EPOCH FROM timestamp) as integer),
                                'bt', bt,
                                'ip', ip,
                                'cc', cc,
                                'vhost', vhost
                            )
                        )
                    FROM
                        attack
                    WHERE (
                        waf_id = %s
                            AND
                        timestamp > to_timestamp(%s)
                    )
                """
                cursor.execute(query, (waf_id, d30_ts,))
            res = cursor.fetchall()

            """
            If the result is not empty, returned a list of tuples, e.g.:
            [([{'bt': 1, ...}, {'bt': 2, ...},...],)]
            """

            # processing result
            if res and res[0][0]:

                # import to pd
                df = pd.DataFrame(res[0][0])

                # del vars
                del res

        except Exception as e:
            log.error(
                'An error occurred while processing data from the database in {}: {}'
                .format(url, e)
            )

        # processing df
        if len(df):

            # processing for each param from the list
            for param in params:

                # replace none
                df.fillna({param: ''}, inplace=True)

                # processing for each unique value of param
                for v in sorted(df[param].unique()):

                    # value normalization
                    if param == 'bt':
                        v = int(v)
                    if param == 'cc':
                        if isinstance(v, str) and (v.strip() == ''):
                            v = None

                    # init
                    ndf = df.loc[(df[param] == v)]

                    # attack by period
                    if use_period:
                        resp[param].append({
                            param: v,
                            'period': len(ndf)
                        })
                    else:
                        resp[param].append({
                            param: v,
                            'lastDay': len(ndf.loc[ndf['timestamp'] > d01_ts]),
                            'last7Days': len(ndf.loc[ndf['timestamp'] > d07_ts]),
                            'last30Days': len(ndf.loc[ndf['timestamp'] > d30_ts])
                        })

            # del vars
            del df

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(json.dumps(resp), status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


##
# Scanning result
##

@app.route(set_scanner_uri, methods=['POST'])
def set_scanner_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # processing data from the request
        try:
            data = request.get_json()
        except Exception as e:
            log.error(
                'An empty or invalid request to {} was received (client IP: {}): {}'
                .format(set_scanner_uri, agent_ip, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=400)

        # processing data from the JSON
        for item in data:

            # decode WAF ID
            try:
                key = base64.b64decode(item['waf_lic_key']).decode('UTF-8', 'ignore')
                if not key:
                    log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
                    return Response(status=401)
            except Exception as e:
                log.error(
                    'An error occurred while processing product key from request to {} (client IP: {}): {}'
                    .format(set_scanner_uri, agent_ip, e)
                )
                return Response(status=401)

            # decode other fields
            domain_scan = base64.b64decode(item['domain']).decode('UTF-8', 'ignore')
            method_scan = base64.b64decode(item['method']).decode('UTF-8', 'ignore')
            param_scan = base64.b64decode(item['param']).decode('UTF-8', 'ignore')
            data_scan = base64.b64decode(item['data']).decode('UTF-8', 'ignore')
            scan_date_scan = base64.b64decode(item['date']).decode('UTF-8', 'ignore')
            payload_scan = base64.b64decode(item['payload']).decode('UTF-8', 'ignore')
            type_scan = base64.b64decode(item['type']).decode('UTF-8', 'ignore')
            url_scan = base64.b64decode(item['URL']).decode('UTF-8', 'ignore')

            # URL decode
            url_scan = unquote(unquote(url_scan))
            param_scan = unquote(unquote(param_scan))

            # mz processing
            if 'mz' in item:
                mz = base64.b64decode(item['mz']).decode('UTF-8', 'ignore')
            else:
                mz = ''

            # severity processing
            if 'severity' in item:
                severity_scan = base64.b64decode(item['severity']).decode('UTF-8', 'ignore')
            else:
                if type_scan.lower() in ['ba', 'lfi', 'rce', 'rfi', 'sqli', 'ssti']:
                    severity_scan = 'High'
                elif type_scan.lower() in ['sde', 'ssrf', 'vc', 'xss']:
                    severity_scan = 'Medium'
                else:
                    severity_scan = 'Low'

            # WAF ID check
            waf_id = get_waf_id(key, agent_ip)
            if not waf_id:
                return Response(status=401)

            ##
            # Storing scanner result to the database
            ##

            # prepare the query
            query = """
                INSERT INTO scan_report (
                    domain, method, scan_date, mz, param, data,
                    payload, type, url, content, waf_id
                ) VALUES (
                    %s, %s, %s, %s, %s, %s,
                    %s, %s, %s, %s, %s
                )
            """

            # prepare the array
            ar = (
                domain_scan, method_scan, scan_date_scan, mz, param_scan, data_scan,
                payload_scan, type_scan, url_scan, severity_scan, waf_id,
            )

            try:

                # insert the data
                cursor.execute(query, ar)

                # logging
                log.info(
                    'Scanner result for domain {} has been successfully stored in the database (client IP: {})'
                    .format(domain_scan, agent_ip)
                )

            except Exception as e:
                log.error(
                    'An error occurred while storing scanner data in the database (client IP: {}): {}'
                    .format(e, agent_ip)
                )
                return Response(status=400)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(status=202)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


##
# Dyn SyncBanIP
##

@app.route(get_dyn_sync_ban_uri, methods=['POST'])
def get_dyn_sync_ban_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Get related keys
        ##

        # init
        relations = []

        try:

            # search the data
            query = """ SELECT key_id FROM dyn_sync_ban WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            res = cursor.fetchall()

            # processing result
            if res:

                # extract relations as list
                relations = [i for sub in res for i in sub]

        except Exception as e:
            log.error(
                'An error occurred while processing relations from the database in {}: {}'
                .format(url, e)
            )

        ##
        # Processing relations
        ##

        # init
        get_ban_data_res = []

        # processing key from relations
        for key_id in relations:
            try:

                # get data from memcached
                memc_key = 'dsbd_' + str(key_id)
                with memc_lock:
                    memc_bin = memc.get(memc_key)

                # processing data
                if memc_bin:

                    # load data
                    data = ast.literal_eval(memc_bin.decode('UTF-8'))

                    # update response result if ret is not empty
                    if data:

                        # data normalization
                        ban_ts = data['timestamp']
                        ban_data = data['ban_data']

                        # add timestamp to ban_data
                        for i, d in enumerate(ban_data.copy()):
                            # set the timestamp
                            d['timestamp'] = ban_ts
                            # update result dictionary
                            get_ban_data_res.append(d)

            except Exception as e:
                log.error(
                    'An error occurred while processing GetDynBanSync data from memcached for key id {}: {}'
                    .format(key_id, e)
                )

        # init
        resp = {'status': 'success', 'ban_data': []}

        # if ban_data exists
        if get_ban_data_res:

            # update response
            resp['ban_data'] = get_ban_data_res

        # convert to JSON
        resp = json.dumps(resp)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_dyn_sync_ban_uri, methods=['POST'])
def set_dyn_sync_ban_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception as e:
            log.error('An error occurred while processing data from {} to {}: {}'.format(agent_ip, url, e))
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            param = 'waf_lic_key'
            if param in jdata:
                key = jdata[param]
            else:
                return Response(status=401)
        except Exception as e:
            log.error(
                'An error occurred while extract product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # convert all keys to lowercase
        jdata = {k.lower(): v for k, v in jdata.items()}

        ##
        # Processing request data
        ##

        # remove unused keys
        jdata = {k: jdata[k] for k in dyn_sync_ban_cols if k in jdata}

        # check columns
        for k in dyn_sync_ban_cols:
            if k not in jdata.keys():
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Required key missing'})
                return Response(resp, status=201)

        # value validation
        for k in dyn_sync_ban_cols:

            # convert to srt and replace newline/multiple space to space
            v = re.sub(r'\s+', ' ', str(jdata[k]))

            # value validations
            status = dyn_sync_ban_value_validation(agent_ip, k, v)

            # check status
            if not status:

                # logging
                log.error('Incorrect value for {} received from {} to {}: {}'.format(k, agent_ip, url, v))

                # response
                close_db_connection(connect, cursor)
                err_msg = str('Incorrect value for {}'.format(k))
                resp = str({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

        # convert key to md5
        key_id = hashlib.md5(key.encode()).hexdigest()

        ##
        # Insert key to relations table
        ##

        try:

            # insert the data
            query = """
                INSERT INTO dyn_sync_ban (
                    key_id, waf_id
                ) VALUES (
                    %s, %s
                ) ON CONFLICT (
                    key_id
                ) DO UPDATE SET
                    waf_id = %s
            """
            cursor.execute(query, (key_id, waf_id, waf_id,))

        except Exception as e:
            log.error('An error occurred while inserting data to dyn_sync_ban: {}'.format(e))
            close_db_connection(connect, cursor)
            return Response(status=502)

        ##
        # Get related keys
        ##

        try:

            # search the data
            query = """ SELECT key_id FROM dyn_sync_ban WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            res = cursor.fetchall()

            # extract relations as list
            relations = [i for sub in res for i in sub]

        except Exception as e:
            log.error(
                'An error occurred while processing relations from the database: {}'
                .format(e)
            )
            close_db_connection(connect, cursor)
            return Response(status=502)

        # receive the data from the memcached if ban_data is not empty
        if jdata['ban_data']:
            try:

                # init
                data = {}

                # get data from memcached
                memc_key = 'dsbd_' + str(key_id)
                with memc_lock:
                    memc_bin = memc.get(memc_key)

                # processing data
                if memc_bin:

                    # load data
                    data = ast.literal_eval(memc_bin.decode('UTF-8'))

                    # add to the data only items with a higher value than the items in the data
                    if data:

                        # extract ban_data
                        api_ban_data = data['ban_data']
                        dyn_ban_data = jdata['ban_data']

                        # processing ban_data from Dyn
                        for item in dyn_ban_data:

                            # skip empty
                            if not item:
                                continue

                            # init
                            is_exists = False

                            # extract IP, TTL and domain
                            ip = item['ip-address']
                            ttl = int(item['TTL'])
                            domain = item['domain']

                            # processing ban_data from API
                            for d in api_ban_data:

                                # search item from dyn_ban_data in api_ban_data
                                if ip == d['ip-address'] and domain == d['domain']:

                                    # if API item's TTL is greater or equal with Dyn item
                                    if int(d['TTL']) >= ttl:
                                        is_exists = True
                                        break

                            # update the result if Dyn item does not exist in API items
                            if not is_exists:
                                data['ban_data'].append(item)

                # use jdata if API data
                if not data:
                    data = jdata

            except Exception as e:
                log.error(
                    'An error occurred while processing SetDynBanSync data for key id {}: {}'
                    .format(key_id, e)
                )
                close_db_connection(connect, cursor)
                return Response(status=502)

        # if ban_data is empty
        else:
            data = jdata

        # init memcached
        memc_key = 'dsbd_' + key_id

        # store data to memcached
        try:
            with memc_lock:
                memc.set(memc_key, str(data).encode(), dyn_sync_ban_memc_ttl)
        except Exception as e:
            log.error(
                'An error occurred while storing dyn_sync_ban data "{}" to memcached from client {}: {} (data: {})'
                .format(memc_key, agent_ip, e, jdata)
            )

        ##
        # Processing relations
        ##

        # init
        get_ban_data_res = []

        # processing key from relations
        for key_id in relations:
            try:

                # get data from memcached
                memc_key = 'dsbd_' + str(key_id)
                with memc_lock:
                    memc_bin = memc.get(memc_key)

                # processing data
                if memc_bin:

                    # load data
                    data = ast.literal_eval(memc_bin.decode('UTF-8'))

                    # update response result if ret is not empty
                    if data:
                        get_ban_data_res.append(json.dumps(data))

            except Exception as e:
                log.error(
                    'An error occurred while processing SetDynBanSync data from memcached for key id {}: {}'
                    .format(key_id, e)
                )

        # init
        resp = json.dumps({'status': 'success'})

        # if ban_data exists
        if get_ban_data_res:
            resp = resp + '\n' + '\n'.join(get_ban_data_res)
        else:
            resp = resp + '\n' + json.dumps(jdata)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(del_dyn_sync_ban_uri, methods=['POST'])
def del_dyn_sync_ban_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception as e:
            log.error('An error occurred while processing data from {} to {}: {}'.format(agent_ip, url, e))
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            param = 'key'
            if param in jdata:
                key = jdata[param]
            else:
                return Response(status=401)
        except Exception as e:
            log.error('An error occurred while extract product key from {} to {}: {}'.format(agent_ip, url, e))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # convert all keys to lowercase
        jdata = {k.lower(): v for k, v in jdata.items()}

        ##
        # Processing request data
        ##

        # init
        param = 'ip-address'

        # extract IP address
        try:

            if param in jdata:

                # extract an IP
                ip = jdata[param].strip()

                ##
                # IP validation
                ##

                # Only IPv4/IPv6/*
                if ip != '*':
                    IPAddress(ip)

            else:

                # response
                close_db_connection(connect, cursor)
                err_msg = str('Key missing: {}'.format(param))
                resp = json.dumps({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

        except Exception as e:
            log.error('An error occurred while processing {} from {} to {}: {}'.format(param, agent_ip, url, e))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # convert key to md5
        key_id = hashlib.md5(key.encode()).hexdigest()

        ##
        # Insert key to relations table
        ##

        try:

            # insert the data
            query = """
                INSERT INTO dyn_sync_ban (
                    key_id, waf_id
                ) VALUES (
                    %s, %s
                ) ON CONFLICT (
                    key_id
                ) DO UPDATE SET
                    waf_id = %s
            """
            cursor.execute(query, (key_id, waf_id, waf_id,))

        except Exception as e:
            log.error('An error occurred while inserting data to dyn_sync_ban: {}'.format(e))
            close_db_connection(connect, cursor)
            return Response(status=502)

        ##
        # Get related keys
        ##

        try:

            # search data
            query = """ SELECT key_id FROM dyn_sync_ban WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            res = cursor.fetchall()

            # extract relations as list
            relations = [i for sub in res for i in sub]

        except Exception as e:
            log.error('An error occurred while processing relations from the database: {}'.format(e))
            close_db_connection(connect, cursor)
            return Response(status=502)

        # close database connection
        close_db_connection(connect, cursor)

        ##
        # Processing relations
        ##

        # processing key from relations
        for key_id in relations:
            try:

                # get data from memcached
                memc_key = 'dsbd_' + str(key_id)
                with memc_lock:
                    memc_bin = memc.get(memc_key)

                # processing data
                if memc_bin:

                    # load data
                    data = ast.literal_eval(memc_bin.decode('UTF-8'))

                    # unban an IP from memcached result
                    if data:

                        # extract ban_data
                        ban_data = data['ban_data']

                        # processing for each IP-address
                        for i, d in enumerate(ban_data):

                            # skip if IP does not match and IP is not a '*'
                            if ip != d['ip-address'] and ip != '*':
                                continue

                            # unban an IP
                            ban_data[i]['is_banned'] = "0"
                            ban_data[i]['TTL'] = str(int(ban_data[i]['TTL']) + 60)

                        # update ban_data
                        data['ban_data'] = ban_data

                    # update data in memcached
                    try:
                        with memc_lock:
                            memc.set(memc_key, str(data).encode(), dyn_sync_ban_memc_ttl)
                    except Exception as e:
                        log.error(
                            'An error occurred while storing dyn_sync_ban data "{}" to memcached from client {}: {} (data: {})'
                            .format(memc_key, agent_ip, e, data)
                        )

            except Exception as e:
                log.error(
                    'An error occurred while processing DelDynBanSync data for key id {}: {}'
                    .format(key_id, e)
                )

        # response
        resp = json.dumps({'status': 'success'})
        return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


##
# OpenAPI
##

@app.route(get_openapi_schemas_uri, methods=['POST'])
def get_openapi_schema_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor(cursor_factory=RealDictCursor)

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # search data
        query = """ SELECT name AS schema_name, convert_from(schema,'UTF8') AS schema FROM openapi_schemas WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'schemas': rows})
        return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(set_openapi_schema_uri, methods=['POST'])
def set_openapi_schema_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)

        # extract the action
        if data is None:
            return Response(status=400)
        elif data.get('set'):
            cmd = 'set'
            schema_data = data.get('set')
        elif data.get('del'):
            cmd = 'del'
            schema_data = data.get('del')
        else:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # extract the schema name
        schema_name = schema_data.get('schema_name')

        # schema name processing
        if not schema_name:
            resp = json.dumps({'status': 'fail', 'description': 'The OpenAPI name of schema is not set'})
            return Response(resp, status=201)

        # create/update the schema
        if cmd == 'set':

            # extract old schema name
            old_name = schema_data.get('old_name')

            # extract the schema
            schema = schema_data.get('schema')

            # if old name and the schema are not set
            if not old_name and not schema:
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': 'Invalid OpenAPI schema'})
                return Response(resp, status=201)

            # schema validation
            if schema:
                try:

                    # load the schema
                    s = json.loads(
                        base64.b64decode(schema).decode('UTF-8')
                    )

                    # schema validation: basic
                    try:
                        validate_spec(s)
                    except Exception as e:
                        close_db_connection(connect, cursor)
                        resp = json.dumps({
                            'status': 'fail',
                            'description': 'Schema validation failed ({})'.format(e)
                        })
                        return Response(resp, status=201)

                    # schema validation: readOnly w/ writeOnly
                    r = oa_validation_rwo(s.copy(), '', 'fail')
                    if r[1] == 'fail':
                        close_db_connection(connect, cursor)
                        resp = json.dumps({
                            'status': 'fail',
                            'description': 'Schema validation failed (readOnly and writeOnly cannot be used at the same time)'
                        })
                        return Response(resp, status=201)

                except Exception:
                    close_db_connection(connect, cursor)
                    resp = json.dumps({'status': 'fail', 'description': 'OpenAPI schema is not in JSON format'})
                    return Response(resp, status=201)

            # search for a schema's name
            query = """ SELECT id FROM openapi_schemas WHERE (waf_id = %s AND name = %s) """

            # search for a schema name
            cursor.execute(query, (waf_id, schema_name))
            rows = cursor.fetchone()

            # search for old schema name
            if old_name:

                # if new schema name already exists
                if rows:
                    close_db_connection(connect, cursor)
                    resp = json.dumps(
                        {'status': 'fail', 'description': 'The OpenAPI schema "{}" already exists'.format(schema_name)}
                    )
                    return Response(resp, status=201)

                # if old schema name does not exist
                cursor.execute(query, (waf_id, old_name))
                if not cursor.fetchone():
                    close_db_connection(connect, cursor)
                    resp = json.dumps(
                        {'status': 'fail', 'description': 'The OpenAPI schema "{}" does not exist'.format(old_name)}
                    )
                    return Response(resp, status=201)

            # if the schema exists or old name is set
            if rows or old_name:

                # if old name is set
                if old_name:

                    if schema:
                        query = """
                            UPDATE
                                openapi_schemas
                            SET
                                schema = %s,
                                timestamp = CURRENT_TIMESTAMP,
                                name = %s
                            WHERE
                                (waf_id = %s AND name = %s)
                        """
                        cursor.execute(query, (schema, schema_name, waf_id, old_name))
                    else:
                        query = """
                            UPDATE
                                openapi_schemas
                            SET
                                timestamp = CURRENT_TIMESTAMP,
                                name = %s
                            WHERE
                                (waf_id = %s AND name = %s)
                        """
                        cursor.execute(query, (schema_name, waf_id, old_name))

                    r = schema_name_after_change(old_name, schema_name, waf_id)
                    if r is False:
                        log.error(
                            'The OpenAPI schema "{}" received from client {} has not been updated'
                            .format(old_name, agent_ip)
                        )
                        close_db_connection(connect, cursor)
                        resp = json.dumps({'status': 'fail', 'description': 'The OpenAPI schema has not been updated'})
                        return Response(resp, status=201)
                    else:
                        log.info(
                            'The OpenAPI schema "{}" received from client {} has been updated'
                            .format(old_name, agent_ip)
                        )

                # if old name is not set
                else:

                    query = """
                        UPDATE
                            openapi_schemas
                        SET
                            schema = %s,
                            timestamp = CURRENT_TIMESTAMP
                        WHERE
                            (waf_id = %s AND name = %s)
                    """
                    cursor.execute(query, (schema, waf_id, schema_name))

                    r = set_openapi_hash('openapi_schemas', waf_id)
                    if r is False:
                        log.error(
                            'The OpenAPI schema "{}" received from client {} has not been updated'
                            .format(schema_name, agent_ip)
                        )
                        close_db_connection(connect, cursor)
                        resp = json.dumps({'status': 'fail'})
                        return Response(resp, status=201)
                    else:
                        log.info(
                            'The OpenAPI schema "{}" received from client {} has been updated'
                            .format(schema_name, agent_ip)
                        )

            # if the schema does not exist
            else:

                # insert the data
                query = """ INSERT INTO openapi_schemas (schema, waf_id, name) VALUES (%s, %s, %s) RETURNING id """
                cursor.execute(query, (schema, waf_id, schema_name))

                r = set_openapi_hash('openapi_schemas', waf_id)
                if r is False:
                    log.error(
                        'The OpenAPI schema "{}" received from client {} has not been created'
                        .format(schema_name, agent_ip)
                    )
                    close_db_connection(connect, cursor)
                    resp = json.dumps({'status': 'fail'})
                    return Response(resp, status=201)
                else:
                    log.info(
                        'The OpenAPI schema with name "{}" received from client {} has been created'
                        .format(schema_name, agent_ip)
                    )

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        # delete the schema
        elif cmd == 'del':

            # delete data
            query = """ DELETE FROM openapi_schemas WHERE (waf_id = %s AND name = %s) """
            cursor.execute(query, (waf_id, schema_name))

            r = set_openapi_hash('openapi_schemas', waf_id)
            if r is False:
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail'})
                return Response(resp, status=201)

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'status': 'fail'})
        return Response(resp, status=201)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_openapi_url_uri, methods=['POST'])
def get_openapi_url_uri_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # search the data
        query = """ SELECT convert_from(url,'UTF8') AS urls FROM openapi_url WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()

        # close database connection
        close_db_connection(connect, cursor)

        # result processing
        if rows:
            try:

                # extract paths list
                rows = json.loads(rows[0])

                # replace any to str
                for idx, row in enumerate(rows):
                    rows[idx] = re.sub(r'{\w+}', '{str}', row)

            except json.JSONDecodeError:
                rows = []

        # response
        resp = json.dumps({'url': rows})
        return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(set_openapi_url_uri, methods=['POST'])
def set_openapi_url_uri_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # load data as list
        schema_url = data.get('url')

        ##
        # IDNA vhost normalization
        ##

        # init
        res = []

        # processing for each schema in schema's list
        for i in schema_url:

            # extract vhost/path
            vhost = i.split('/', 1)[0]
            path = '/' + i.split('/', 1)[1]

            # non-latin vhost normalization
            vhost = idna_vhost_list_normalization(vhost)

            # update list
            res.append(vhost + path)

        # update schema_url
        schema_url = json.dumps(res.copy())

        # processing data
        if schema_url:

            if not schema_url.strip('[]'):

                query = """ DELETE FROM openapi_url WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # delete data from memcached
                memc_key = 'hash_url_' + waf_id
                with memc_lock:
                    memc.delete(memc_key)

            else:

                query = """ SELECT id FROM openapi_url WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))
                rows = cursor.fetchone()

                if rows:

                    # update data
                    query = """ UPDATE openapi_url SET url = %s, timestamp = CURRENT_TIMESTAMP WHERE (waf_id = %s) """
                    cursor.execute(query, (schema_url, waf_id))

                    r = set_openapi_hash('openapi_url', waf_id)
                    if r is False:
                        close_db_connection(connect, cursor)
                        resp = json.dumps({'status': 'fail'})
                        return Response(resp, status=201)

                else:

                    # insert the data
                    query = """ INSERT INTO openapi_url (url, waf_id) VALUES (%s, %s) RETURNING id """
                    cursor.execute(query, (schema_url, waf_id))

                    r = set_openapi_hash('openapi_url', waf_id)
                    if r is False:
                        close_db_connection(connect, cursor)
                        resp = json.dumps({'status': 'fail'})
                        return Response(resp, status=201)

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'status': 'fail'})
        return Response(resp, status=201)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(set_openapi_schema_paths_uri, methods=['POST'])
def set_openapi_schema_paths_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # extract the action
        set_path = data.get('set')
        del_path = data.get('del')

        # extract the key
        if set_path:
            key = set_path.get('key')
        else:
            key = del_path.get('key')

        # key processing
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # update/insert the data
        if set_path:

            # extract the request data
            schema_name = set_path.get('schema_name')
            schema_url = set_path.get('url')
            schema = set_path.get('schema')

            # data validation
            if schema_name is None:
                schema_name = del_path.get('name')
            if not schema_name:
                resp = json.dumps({'status': 'fail', 'description': 'An empty schema name'})
                close_db_connection(connect, cursor)
                return Response(resp, status=201)

            # search data
            query = """ SELECT id FROM openapi_schema_paths WHERE (schema_name = %s AND url = %s AND waf_id = %s) """
            cursor.execute(query, (schema_name, schema_url, waf_id))
            rows = cursor.fetchone()

            # update data
            if rows:

                query = """
                    UPDATE
                        openapi_schema_paths
                    SET
                        schema_name = %s,
                        url = %s,
                        schema = %s,
                        timestamp = CURRENT_TIMESTAMP
                    WHERE
                        (schema_name = %s AND url = %s AND waf_id = %s)
                """
                cursor.execute(query, (schema_name, schema_url, schema, schema_name, schema_url, waf_id))

            # insert data
            else:

                query = """
                    INSERT INTO openapi_schema_paths (
                        schema_name, schema, url, waf_id
                    ) VALUES (
                        %s, %s, %s, %s
                    )
                """
                cursor.execute(query, (schema_name, schema, schema_url, waf_id))

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        # delete the data
        elif del_path:

            # extract the request data
            schema_name = del_path.get('schema_name')
            schema_url = del_path.get('url')

            # data validation
            if schema_name is None:
                schema_name = del_path.get('name')
            if not schema_name:
                resp = json.dumps({'status': 'fail', 'description': 'An empty schema name'})
                return Response(resp, status=201)

            query = """ DELETE FROM openapi_schema_paths WHERE (url = %s AND schema_name = %s AND waf_id = %s) """
            cursor.execute(query, (schema_url, schema_name, waf_id))

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'status': 'fail'})
        return Response(resp, status=201)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_openapi_schema_paths_uri, methods=['POST'])
def get_openapi_schema_paths_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # init
        result = []

        # receive the data from the database
        query = """ SELECT schema_name, url, convert_from(schema,'UTF8') AS schema FROM openapi_schema_paths WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # data processing
        for row in res:

            # skip empty
            if not row:
                continue

            # update the result
            result.append({
                'schema_name': row[0],
                'url': row[1],
                'schema': row[2]
            })

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'paths': result})
        return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_openapi_exclude_paths_uri, methods=['POST'])
def get_openapi_exclude_paths_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            close_db_connection(connect, cursor)
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        query = """ SELECT convert_from(exclude_paths,'UTF8') AS exclude_paths FROM openapi_exclude_paths WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()

        if rows:
            try:
                rows = json.loads(rows[0])
            except json.JSONDecodeError:
                rows = rows

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'exclude_paths': rows})
        return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(set_openapi_exclude_paths_uri, methods=['POST'])
def set_openapi_exclude_paths_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            close_db_connection(connect, cursor)
            return Response(status=400)

        set_exclude_paths = data.get('set')

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        if set_exclude_paths:
            exclude_paths = set_exclude_paths.get('exclude_paths')
            if exclude_paths is None:
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail'})
                return Response(resp, status=201)
            else:
                exclude_paths = json.dumps(exclude_paths)

            query = """ SELECT id FROM openapi_exclude_paths WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))
            rows = cursor.fetchone()

            if not exclude_paths:

                query = """ DELETE FROM openapi_exclude_paths WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # get data from memcached
                memc_key = 'hash_exclude_path_' + waf_id
                with memc_lock:
                    memc.delete(memc_key)

            else:

                if rows:

                    query = """
                        UPDATE
                            openapi_exclude_paths
                        SET
                            exclude_paths = %s,
                            timestamp = CURRENT_TIMESTAMP
                        WHERE
                            (waf_id = %s)
                    """
                    cursor.execute(query, (exclude_paths, waf_id))

                    r = set_openapi_hash('openapi_exclude_paths', waf_id)
                    if r is False:
                        close_db_connection(connect, cursor)
                        resp = json.dumps({'status': 'fail'})
                        return Response(resp, status=201)

                else:

                    query = """
                        INSERT INTO openapi_exclude_paths (
                            exclude_paths, waf_id
                        ) VALUES (
                            %s, %s
                        ) RETURNING
                            id
                    """
                    cursor.execute(query, (exclude_paths, waf_id))

                    r = set_openapi_hash('openapi_exclude_paths', waf_id)
                    if r is False:
                        close_db_connection(connect, cursor)
                        resp = json.dumps({'status': 'fail'})
                        return Response(resp, status=201)

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'status': 'fail'})
        return Response(resp, status=201)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_openapi_exclude_paths_hash_uri, methods=['POST'])
def get_openapi_exclude_paths_hash_f():

    # set URL
    url = request.path

    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return Response(status=401)

        # get data from memcached
        memc_key = 'hash_exclude_path_' + waf_id
        with memc_lock:
            memc_bin = memc.get(memc_key)

        # processing data
        if memc_bin:
            resp = "{{'status': 'success', 'hash': '{hash}'}}".format(hash=memc_bin.decode('UTF-8'))
            return Response(resp, status=200)
        else:
            resp = "{'status': 'none'}"
            return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_openapi_schemas_hash_uri, methods=['POST'])
def get_openapi_schema_hash_f():

    # set URL
    url = request.path

    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return Response(status=401)

        # get data from memcached
        memc_key = 'hash_schemas_' + waf_id
        with memc_lock:
            memc_bin = memc.get(memc_key)

        # processing data
        if memc_bin:
            resp = "{{'status': 'success', 'hash': '{hash}'}}".format(hash=memc_bin.decode('UTF-8'))
            return Response(resp, status=200)
        else:
            resp = "{'status': 'none'}"
            return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_openapi_url_hash_uri, methods=['POST'])
def get_openapi_url_hash_f():

    # set URL
    url = request.path

    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract the data
        data = get_data_from_json(request, url)
        if data is None:
            return Response(status=400)

        # key processing
        key = data.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return Response(status=401)

        # get data from memcached
        memc_key = 'hash_url_' + waf_id
        with memc_lock:
            memc_bin = memc.get(memc_key)

        # processing data
        if memc_bin:
            resp = "{{'status': 'success', 'hash': '{hash}'}}".format(hash=memc_bin.decode('UTF-8'))
            return Response(resp, status=200)
        else:
            resp = "{'status': 'none'}"
            return Response(resp, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


##
# Models tuning
##

@app.route(set_mt_data_uri, methods=['POST'])
def set_mt_data_uri_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        try:
            agent = request.user_agent.string
        except Exception:
            agent = ''

        try:
            data = base64.b64decode(request.get_data()).decode('UTF-8', 'ignore')
        except Exception as e:
            log.error('An error occurred while processing the data in {}: {}'.format(url, e))
            close_db_connection(connect, cursor)
            return Response(status=202)

        try:
            jdata = json.loads(data)
        except Exception as e:
            log.error('An error occurred while converting data JSON in {}: {} ({})'.format(url, e, data))
            close_db_connection(connect, cursor)
            return Response(status=202)

        # del vars
        del data

        # extract lic. key
        key = request.cookies['key']
        if not key:
            log.error(
                'An error occurred while processing data in {} from client {} ({}): no product key'
                .format(url, agent_ip, agent)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # make record unique
        # jdata = [dict(y) for y in set(tuple(x.items()) for x in jdata)]

        # processing for each item
        for i in range(len(jdata)):

            # set WAF ID
            jdata[i]['waf_id'] = waf_id

            # timestamp processing and data normalization
            if len(str(jdata[i]['timestamp'])) > 10:
                jdata[i]['timestamp'] = str(jdata[i]['timestamp'])[:10]
            try:
                datetime.fromtimestamp(int(jdata[i]['timestamp'])).strftime('%H:%M:%S %d.%m.%Y')
            except Exception as e:
                log.error(
                    'An error occurred in {} while converting date {}: {}'.format(url, jdata[i]['timestamp'], e)
                )
                return Response(status=202)

            # add agent field
            if 'MLA' in agent:
                jdata[i]['agent'] = 'MLA'
            elif 'MLC' in agent:
                jdata[i]['agent'] = 'MLC'
            else:
                jdata[i]['agent'] = '-'

            # data normalization
            for z in ['uri', 'args', 'cookie', 'referer', 'other_headers', 'body', 'ua']:
                if isinstance(jdata[i][z], str):
                    jdata[i][z] = jdata[i][z].replace('\x00', '\\x00')

            ##
            # Zone normalization
            ##

            # convert to lowercase
            jdata[i]['mz'] = str(jdata[i]['mz']).lower()

            # init
            r = []

            # rename zones
            for mz in jdata[i]['mz'].split('|'):

                # rename the zone
                if mz == 'url':
                    mz = 'uri'
                elif mz == 'user-agent':
                    mz = 'ua'

                # update the list
                r.append(mz)

            # update zones list
            jdata[i]['mz'] = '|'.join(r)
            del r

            ##
            # Save data to the database
            ##

            try:

                # set the fields
                fields_list = (
                    'bt', 'timestamp', 'mz', 'ip', 'vhost', 'method',
                    'uri', 'args', 'cookie', 'referer', 'other_headers', 'body',
                    'mz_origin', 'ua', 'request_id', 'waf_id', 'agent'
                )

                query = """
                    INSERT INTO
                        mt_main
                    (
                        {}
                    ) VALUES (
                        %s, %s, %s, %s, %s, %s,
                        %s, %s, %s, %s, %s, %s,
                        %s, %s, %s, %s, %s
                    )
                """.format(','.join(fields_list))

                arr = (jdata[i].get(field) for field in fields_list)
                cursor.execute(query, tuple(arr))

            except Exception as e:
                log.error('An error occurred while storing data into mt_main: {} ({})'.format(e, jdata[i]))
                close_db_connection(connect, cursor)
                return Response(status=202)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(status=202)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=202)


@app.route(get_mt_data_uri, methods=['GET'])
def get_mt_export_f():

    # set URL
    url = request.path

    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.cookies['key']
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return Response(status=401)

        res = mt_get_export(waf_id)
        if res:
            d = json.dumps(res)
            resp = base64.b64encode(bytes(d, 'UTF-8'))
            return Response(resp, status=200)
        else:
            return Response(status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


@app.route(get_mt_hash_uri, methods=['GET'])
def get_mt_timestamp_f():

    # set URL
    url = request.path

    try:

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.cookies['key']
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return Response(status=401)

        res = mt_get_export_md5(waf_id)
        return Response(res, status=200)

    except Exception as e:
        log.exception('An error occurred while processing the request to {}: {}'.format(url, e))
        return Response(status=400)


##
# Models management
##

@app.route(get_models_uri, methods=['POST'])
def get_models_uri_f():

    """
    Status code:
    - 202 : the model exists
    - 203 : the model does not exist
    """

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract key
        key = request.cookies.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        try:

            # processing data from the request
            data = base64.b64decode(request.data).decode('UTF-8')

            # check data from the request
            if not data:
                close_db_connection(connect, cursor)
                return Response(status=401)

            # load models list from agent
            a_models_dict = json.loads(data)

            # if check status request
            if 'is_active' in a_models_dict.keys():
                close_db_connection(connect, cursor)
                resp = json.dumps({'is_active': True})
                return Response(resp, 200)

        except Exception as e:
            log.error('An error occurred in {} while processing data from the request: {}'.format(url, e))
            return Response(status=400)

        ##
        # Get models list from the database
        ##

        # init
        l_models_dict = {}

        # receiving vhosts and hashes for all models by WAF ID
        query = """ SELECT vhost, hash FROM models WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # generate db models dict
        for row in res:
            vhost = str(row[0])
            shash = str(row[1])
            l_models_dict[vhost] = shash

        try:

            ##
            # Convert dict to list
            ##

            # list from remote agent
            a_models_list = []
            for k in a_models_dict.keys():
                k = idna_vhost_list_normalization(k)
                a_models_list.append(str(k) + ':' + str(a_models_dict[k]))

            # list from the database
            l_models_list = []
            for k in l_models_dict.keys():
                k = idna_vhost_list_normalization(k)
                l_models_list.append(str(k) + ':' + str(l_models_dict[k]))

            # if models exists in the database
            if l_models_dict:

                # get models list, that the remote client does not have
                get_models_list = list(set(l_models_list) - set(a_models_list))

                # if unique models list is not empty, get models for first list element by hash
                if get_models_list:

                    # init
                    vhost = str(get_models_list[0].split(':')[0])
                    shash = str(get_models_list[0].split(':')[1])

                    # get models by vhost and hash
                    query = """ SELECT model FROM models WHERE (vhost = %s AND hash = %s AND waf_id = %s) """
                    cursor.execute(query, (vhost, shash, waf_id,))
                    res = cursor.fetchone()
                    close_db_connection(connect, cursor)

                    # if models exist
                    if res:
                        return Response(base64.b64encode(res[0]), status=202)
                    else:
                        return Response(status=304)

                # if the all models from the database are already on the remote side (MLA/MLC)
                else:

                    # close database connection
                    close_db_connection(connect, cursor)

                    # get unique models list, that the db does not have
                    del_models_list = list(set(a_models_list) - set(l_models_list))

                    # response
                    if len(del_models_list):
                        return Response(str(del_models_list), status=203)
                    else:
                        return Response(status=304)

            # models don't exist in the database
            else:

                # close database connection
                close_db_connection(connect, cursor)

                # get unique models list, that the db does not have
                del_models_list = list(set(a_models_list) - set(l_models_list))

                # if there are models on a remote server. (MLA/MLC)
                if a_models_dict:
                    return Response(str(del_models_list), status=203)
                else:
                    return Response(status=304)

        except Exception as e:
            log.error('An error occurred in {}: {}'.format(url, e))
            close_db_connection(connect, cursor)
            return Response(status=400)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))
        return Response(status=503)


@app.route(set_models_uri, methods=['POST'])
def set_models_uri_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract key
        key = request.cookies.get('key')
        if not key:
            log.error('An error occurred while processing data in {} from client {}: no product key'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        try:

            # extract data
            data = json.loads(request.data)
            vhost = data['vhost']
            shash = data['hash']
            model = base64.b64decode(data['model'])
            del data

        except Exception as e:
            log.error('An error occurred in {} while processing data of the request: {}'.format(url, e))
            close_db_connection(connect, cursor)
            return Response(status=400)

        # if models in the request:
        if model:
            try:

                # idna normalization
                vhost = idna_vhost_list_normalization(vhost)

                # search models
                query = """ SELECT vhost FROM models WHERE (vhost = %s AND waf_id = %s) """
                cursor.execute(query, (vhost, waf_id,))
                res = cursor.fetchone()

                # if models are existed
                if res:

                    # update data
                    query = """
                        UPDATE
                            models
                        SET
                            (vhost, waf_id, model, hash) = (%s, %s, %s, %s)
                        WHERE
                            (vhost = %s AND waf_id = %s)
                    """
                    cursor.execute(query, (vhost, waf_id, psycopg2.Binary(model), shash, vhost, waf_id,))

                    # logging
                    log.info('Models for {} with WAF ID {} received from client {}'.format(vhost, waf_id, agent_ip,))

                # if models doesn't exists
                else:

                    # insert the data
                    query = """ INSERT INTO models (vhost, waf_id, model, hash) VALUES (%s, %s, %s, %s) """
                    cursor.execute(query, (vhost, waf_id, psycopg2.Binary(model), shash,))

                    # logging
                    log.info('Models for {} with WAF ID {} received from client {}'.format(vhost, waf_id, agent_ip,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                return Response(status=202)

            except Exception as e:
                log.error(
                    'An error occurred in {} while updating/storing models'
                    ' in the database received from client {}: {}'
                    .format(url, agent_ip, e)
                )
                close_db_connection(connect, cursor)
                return Response(status=500)

        else:

            # close database connection
            close_db_connection(connect, cursor)

            # response
            return Response(status=401)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))
        return Response(status=503)


@app.route(del_models_uri, methods=['POST'])
def del_models_uri_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # extract vhost
        vhost = request.values.get('vhost')
        if vhost is None:
            log.error('An error occurred in {} from {}: no vhost'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        ##
        # Delete models by vhost + WAF ID
        ##

        try:

            # idna normalization
            vhost = idna_vhost_list_normalization(vhost)

            # vhost normalization
            vhost = vhost_normalization(vhost)

            # delete models
            query = """ DELETE FROM models WHERE (vhost = %s AND waf_id = %s) """
            cursor.execute(query, (vhost, waf_id,))

            # logging
            log.info('The model of virtual host {} has been deleted'.format(vhost))

            # close database connection
            close_db_connection(connect, cursor)

            # processing result
            resp = str({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:
            log.error('An error occurred in {} while deleting data: {}'.format(url, e))

        # Default response
        close_db_connection(connect, cursor)
        return Response(status=400)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))
        return Response(status=503)


@app.route(rep_models_uri, methods=['POST'])
def rep_models_uri_f():

    # set URL
    url = request.path

    """
    The wld-vhost for dst-model must be normalized to all, sub_example.com or .example.com.
    The files of wld-vhost for dst-model must be normalized to all, sub_example.com or example.com.
    """

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # extract source models vhost
        src = request.values.get('src')
        if src is None:
            log.error('An error occurred in {} from {}: no src'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        # extract dst models vhost
        dst = request.values.get('dst')
        if dst is None:
            log.error('An error occurred in {} from {}: no dst'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        # idna normalization
        src = idna_vhost_list_normalization(src)
        dst = idna_vhost_list_normalization(dst)

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(dst, waf_id)
        if not is_vhost_exists:
            resp = str({'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'})
            close_db_connection(connect, cursor)
            return Response(resp, status=201)

        ##
        # Copy vhost to dst models with the same WAF ID
        ##

        # vhost normalization
        dst = vhost_normalization(dst)

        try:

            # logging
            log.info(
                'Request to copy models for {} to {} with WAF ID {} received from client {}'
                .format(src, dst, waf_id, agent_ip)
            )

            # search models
            query = """ SELECT model, hash, timestamp FROM models WHERE (vhost = %s AND waf_id = %s) """
            cursor.execute(query, (src, waf_id,))
            res = cursor.fetchone()

            # processing result
            if not res:
                resp = str({'status': 'fail', 'description': 'No models found'})
                close_db_connection(connect, cursor)
                return Response(resp, status=201)

            ##
            # Update filenames in models.zip
            ##

            # init
            smodels = res[0]
            shash = res[1]
            sts = res[2]
            del res

            # init models path
            tmp_dir = os.path.join('/tmp', str(time_ns()))
            smodels_path = os.path.join(tmp_dir, 'smodels.zip')
            dmodels_path = os.path.join(tmp_dir, 'dmodels.zip')

            # create temporary directory
            if not os.path.exists(tmp_dir):
                os.mkdir(tmp_dir)

            # save src models
            with open(smodels_path, 'wb') as f:
                f.write(smodels)

            # init
            smodels = ZipFile(smodels_path, 'r')
            dmodels = ZipFile(dmodels_path, 'w')

            # update model's files
            for file in smodels.filelist:
                # save files
                if file.filename.endswith('_rf.ml'):
                    dmodels.writestr(dst + '_rf.ml', smodels.read(file.filename))
                if file.filename.endswith('_vect.ml'):
                    dmodels.writestr(dst + '_vect.ml', smodels.read(file.filename))

            # close file
            smodels.close()
            dmodels.close()

            # load dst models
            with open(dmodels_path, 'rb') as f:
                dmodels = f.read()

            # delete previous models
            query = """ DELETE FROM models WHERE (vhost = %s AND waf_id = %s) """
            cursor.execute(query, (dst, waf_id))

            # insert new models
            query = """
                INSERT INTO models (
                    vhost, waf_id, model, hash, timestamp
                ) VALUES (
                    %s, %s, %s, %s, %s
                )
            """
            cursor.execute(query, (dst, waf_id, psycopg2.Binary(dmodels), shash, sts))

            # logging
            log.info('The model of virtual host {} is created from the model of virtual host {}'.format(dst, src))

            # del vars
            del smodels, dmodels

            # del tmp_dir
            if os.path.exists(tmp_dir):
                shutil.rmtree(tmp_dir)

            ##
            # Update training status
            ##

            # init
            duration = False
            complete = True
            progress = 100

            # get training status
            query = """ SELECT duration, complete, progress FROM training WHERE (vhost = %s AND waf_id = %s) """
            cursor.execute(query, (src, waf_id,))
            res = cursor.fetchone()

            # processing result
            if res:
                duration = res[0]
                complete = res[1]
                progress = res[2]

            # insert the data
            query = """
                INSERT INTO
                    training (duration, complete, progress, vhost, waf_id)
                VALUES
                    (%s, %s, %s, %s, %s)
                ON CONFLICT
                    (vhost, waf_id)
                DO UPDATE SET
                    (duration, complete, progress) = (%s, %s, %s)
            """
            cursor.execute(query, (duration, complete, progress, dst, waf_id, duration, complete, progress,))

            # del vars
            del res

            # close database connection
            close_db_connection(connect, cursor)

            # processing result
            resp = str({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:
            log.error('An error occurred in {} while copying models: {}'.format(url, e))

        # default response
        close_db_connection(connect, cursor)
        return Response(status=400)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))
        return Response(status=503)


@app.route(get_models_list_uri, methods=['POST'])
def get_models_list_uri_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        try:

            # receiving vhost for all models by waf_id
            query = """ SELECT vhost FROM models WHERE (waf_id = %s) """
            cursor.execute(query, (waf_id,))

            # get models list as list
            result = [i[0] for i in cursor.fetchall()]

            # model's name normalization
            for i, r in enumerate(result.copy()):
                if r == 'all':
                    result[i] = '*'
                elif r.startswith('sub_'):
                    result[i] = r.replace('sub_', '*.')

            # sort result
            result.sort()

            # processing result
            result = '[]' if not result else str(result)

            # close database connection
            close_db_connection(connect, cursor)

            # response
            return Response(result, status=200)

        except Exception as e:
            log.error('An error occurred in {} while processing data: {}'.format(url, e))
            close_db_connection(connect, cursor)
            return Response(status=503)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=400)


@app.route(get_training_uri, methods=['POST'])
def get_training_uri_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, _ = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # extract vhost
        vhost = request.values.get('vhost')
        if key is None:
            log.error('An error occurred in {} from {}: no vhost'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        ##
        # Extract training info of vhost from the database
        ##

        # get training status
        query = """ SELECT duration, complete, progress FROM training WHERE (vhost = %s AND waf_id = %s) """
        cursor.execute(query, (vhost, waf_id,))
        res = cursor.fetchone()

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res:
            resp = str({'status': 'none'})
            close_db_connection(connect, cursor)
            return Response(resp, status=200)
        else:

            # processing duration (default 4 days)
            duration = 4 if res[0] is None else res[0]

            # processing complete/progress
            complete = ast.literal_eval(str(res[1]))
            progress = 0 if res[2] is None else res[2]

            # response
            close_db_connection(connect, cursor)
            resp = str(
                {'status': 'success', 'duration': duration, 'complete': complete, 'progress': progress}
            )
            return Response(resp, status=200)

    except Exception as e:
        log.error(
            'An error occurred in {}: {}'
            .format(url, e)
        )
        return Response(status=400)


@app.route(set_training_uri, methods=['POST'])
def set_training_uri_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # Init training data
        # for further database processing
        training_data = {}

        # extract vhost
        vhost = request.values.get('vhost')
        if key is None:
            log.error('An error occurred in {} from {}: no vhost'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        # update training_data
        training_data['vhost'] = vhost
        training_data['waf_id'] = waf_id

        ##
        # Extract training status/duration
        ##

        # init
        required_keys = ['complete', 'duration']

        # extract data
        for k in required_keys:
            try:

                # extract data
                v = request.values.get(k)

                # skip empty
                if v is None:
                    continue

                # data validation
                if k == 'duration':

                    try:
                        duration = int(v)
                    except Exception:
                        close_db_connection(connect, cursor)
                        resp = str({'status': 'fail', 'description': 'Parameter "duration" is not an integer'})
                        return Response(resp, status=201)

                    # check duration
                    if duration < 1:
                        close_db_connection(connect, cursor)
                        resp = str({'status': 'fail', 'description': 'Parameter "duration" cannot be less than 1'})
                        return Response(resp, status=201)

                # update training_data if value is not empty
                training_data[k] = v

            except Exception:
                pass

        ##
        # Check required keys in training_data
        ##

        if not [k for k in list(training_data.keys()) if k in required_keys]:
            log.error(
                'An error occurred in {} from {}: complete and duration are missing'.
                format(url, agent_ip,)
            )
            close_db_connection(connect, cursor)
            return Response(status=400)

        ##
        # vhost normalization
        ##

        # idna normalization
        vhost = idna_vhost_list_normalization(vhost)

        # all
        if vhost == 'all':
            vhost = '*'

        # *.example.com
        if re.search(r'^sub_', vhost):
            vhost = re.sub(r'^sub_', '*.', vhost)

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(vhost, waf_id)
        if not is_vhost_exists:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'})
            return Response(resp, status=201)

        ##
        # Insert/update data in the database
        ##

        try:

            # delete keys with None value
            training_data = {k: v for k, v in training_data.items() if v is not None}

            # extract data as tuple
            k = ','.join(training_data.keys())
            v = tuple(training_data.values())

            # insert the data
            query = """ INSERT INTO training ({}) VALUES %s ON CONFLICT (vhost, waf_id) DO UPDATE SET ({}) = %s """.format(k, k)
            cursor.execute(query, (v, v,))

            # close database connection
            close_db_connection(connect, cursor)

            # response
            resp = str({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:
            log.error('An error occurred in {} while storing data: {}'.format(url, e))
            close_db_connection(connect, cursor)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # Default response
    return Response(status=400)


@app.route(set_training_progress_uri, methods=['POST'])
def set_training_progress_uri_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # Init training data
        # for further database processing
        training_data = {}

        # extract vhost
        vhost = request.values.get('vhost')
        if vhost is None:
            log.error('An error occurred in {} from {}: no vhost'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        # extract progress
        progress = request.values.get('progress')
        if progress is None:
            log.error('An error occurred in {} from {}: no progress'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=400)

        # update training_data
        training_data['vhost'] = vhost
        training_data['waf_id'] = waf_id
        training_data['progress'] = progress

        ##
        # vhost normalization
        ##

        # idna normalization
        vhost = idna_vhost_list_normalization(vhost)

        # all
        if vhost == 'all':
            vhost = '*'

        # *.example.com
        if re.search(r'^sub_', vhost):
            vhost = re.sub(r'^sub_', '*.', vhost)

        ##
        # Processing vhost status
        ##

        is_vhost_exists = training_vhost_processing(vhost, waf_id)
        if not is_vhost_exists:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The virtual host was not found, set virtual host using Cabinet/API'})
            return Response(resp, status=201)

        ##
        # Insert/update data in the database
        ##

        try:

            query = """ UPDATE training SET progress = %s WHERE (vhost = %s AND waf_id = %s) """
            cursor.execute(query, (progress, vhost, waf_id,))

            # close database connection
            close_db_connection(connect, cursor)

            # response
            resp = str({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:
            log.error('An error occurred in {} while storing data: {}'.format(url, e))

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # Default response
    return Response(status=400)


##
# ML Virtual Hosts Management
##

@app.route(get_vhosts_list_uri, methods=['POST'])
def get_vhosts_list_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, _ = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        query = """ SELECT vhosts_list FROM vhosts_list WHERE (waf_id = %s) LIMIT 1 """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()

        # init
        resp = {'status': 'success'}

        # processing result
        if not res:
            resp['vhosts_list'] = '[]'
        else:
            resp['vhosts_list'] = str(res[0])

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(str(resp), status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_vhosts_list_uri, methods=['POST'])
def set_vhosts_list_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error('An error occurred while extracting product key from {} to {}: {}'.format(agent_ip, url, e))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # Business/Enterprise with models
        if not waf_id or not tarif_id or not models_num:
            return Response(status=401)

        # extract vhosts_list
        try:

            # extract value
            vhosts_list = jdata['vhosts_list']

            # vhosts_list size validation
            if len(vhosts_list) > models_num:
                desc = (
                    'The number of virtual hosts ({}) exceeds the number of available models ({}) for your tariff plan'
                    .format(len(vhosts_list), models_num)
                )
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': desc})
                return Response(resp, status=201)

            # block using example.com with .example.com
            res = [x for x in vhosts_list if '.' + x in vhosts_list]
            if res:
                desc = 'Simultaneous use of virtual host {} and .{} is prohibited'.format(res[0], res[0])
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': desc})
                return Response(resp, status=201)

            ##
            # vhosts_list normalization
            ##

            # init
            nl = []

            # processing vhosts_list
            for v in vhosts_list:

                # skip empty
                if not v:
                    continue

                # idna normalization
                v = idna_vhost_list_normalization(v)

                # *
                if v == 'all':
                    nl.append('*')

                # *.example.com
                elif re.search(r'^sub_', v):
                    nl.append(re.sub(r'^sub_', '*.', v))

                else:
                    nl.append(v)

            # update vhosts_list
            vhosts_list = nl

            # sort by name
            vhosts_list = sorted(vhosts_list)

            # then sort list by length
            vhosts_list.sort(key=len, reverse=True)

            # replace single quotes
            vhosts_list = str(vhosts_list).replace("'", '"')

        except Exception as e:
            log.error(
                'An error occurred while processing vhosts_list from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            return Response(status=401)

        ##
        # Save vhosts_list to the database
        ##

        try:

            query = """
                INSERT INTO vhosts_list (
                    vhosts_list, waf_id
                ) VALUES (
                    %s, %s
                ) ON CONFLICT (
                    waf_id
                ) DO UPDATE SET
                    vhosts_list = %s
            """
            cursor.execute(query, (vhosts_list, waf_id, vhosts_list,))

            # close database connection
            close_db_connection(connect, cursor)

            # response
            resp = str({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:

            # response
            resp = str({'status': 'fail', 'description': str(e)})
            return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


##
# Agent settings
##

@app.route(get_mlc_settings_hash_uri, methods=['POST'])
def get_mlc_settings_hash_f():

    # set URL
    url = request.path

    # extract client IP
    agent_ip = get_agent_ip(request)

    # extract lic. key
    key = request.values.get('key')
    if not key:
        log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
        return Response(status=401)

    # WAF ID check
    waf_id = get_waf_id(key, agent_ip)
    if not waf_id:
        return Response(status=401)

    # processing request
    t = 'mlc_settings'
    resp, resp_code = get_hash(agent_ip, url, t, waf_id)

    # response
    if resp_code == 200:
        return Response(resp, status=resp_code)
    else:
        return Response(status=resp_code)


@app.route(get_mlc_settings_uri, methods=['POST'])
def get_mlc_settings_f():

    # set URL
    url = request.path

    try:

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.args.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        query = """ SELECT to_jsonb(array_agg(mlc_settings)) FROM mlc_settings WHERE (waf_id = %s AND active IS TRUE) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res:

            # init
            resp = {'status': 'none'}

            # format result
            if resp_in_json:
                resp = json.dumps(resp)
            else:
                resp = str(resp)

            # response
            return Response(resp, status=200)

        else:

            # extract first item from list
            # (only 1 item must be received from the database)
            res = res[0]

            # set status
            res['status'] = 'success'

            # del columns
            del res['active']
            del res['modify']
            del res['waf_id']

            # delete keys with None/empty value
            res = {k: v for k, v in res.items() if v is not None and v != '' and v != '[]'}

            # format result
            if resp_in_json:
                resp = json.dumps(res)
            else:
                resp = str(res)

            # response
            return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_mlc_settings_uri, methods=['POST'])
def set_mlc_settings_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, _, _ = get_auth(key, agent_ip)

        # WAF ID check
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in jdata:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = mlc_settings_cols + ['active']

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            ##
            # Create row if it does not exist
            ##

            try:

                # insert the data
                query = """ INSERT INTO mlc_settings (waf_id) VALUES (%s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (waf_id,))

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

            ##
            # Processing request data
            ##

            for k in cols_ext:

                # processing data
                if k in jdata[cmd]:

                    # processing items that not allowed for Light
                    if not tarif_id and k in mlc_be_cols:

                        # response
                        err_msg = 'Parameter "{}" is not available for your plan'.format(k)
                        close_db_connection(connect, cursor)
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # value validations
                    status = mlc_value_validation(agent_ip, k, jdata[cmd][k])

                    # check status
                    if not status:

                        # response
                        err_msg = str('Incorrect value for {}'.format(k))
                        close_db_connection(connect, cursor)
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # replace quotes in list
                    if isinstance(jdata[cmd][k], list):
                        v = str(jdata[cmd][k]).replace("'", '"')
                    else:
                        v = jdata[cmd][k]

                    # convert to srt and replace newline/multiple space to space
                    if isinstance(v, str):
                        v = re.sub(r'\s+', ' ', v)

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            ##
            # Save data to the database
            ##

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                return Response(resp, status=201)

            try:

                try:

                    # prepare SQL data
                    k_list.extend(['modify'])
                    v_list.extend([str(datetime.now())])

                    # update data
                    query = """ UPDATE mlc_settings SET ({}) = %s WHERE (waf_id = %s) """.format(','.join(k_list))
                    cursor.execute(query, (tuple(v_list), waf_id,))

                except Exception as e:

                    # logging
                    log_key = 'x' * 26 + key[-6:]
                    log.info('An error occurred while processing data for key "{}" in mlc_settings: {}'.format(log_key, e))

                    # response
                    close_db_connection(connect, cursor)
                    resp = str({'status': 'fail', 'description': str(e)})
                    return Response(resp, status=201)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

        # processing command
        elif cmd == 'del':

            ##
            # Delete settings from the database
            ##

            try:

                query = """ DELETE FROM mlc_settings WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_dyn_settings_hash_uri, methods=['POST'])
def get_dyn_settings_hash_f():

    # set URL
    url = request.path

    # extract client IP
    agent_ip = get_agent_ip(request)

    # extract lic. key
    key = request.values.get('key')
    if not key:
        log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
        return Response(status=401)

    # WAF ID check
    waf_id = get_waf_id(key, agent_ip)
    if not waf_id:
        return Response(status=401)

    # processing request
    t = 'dyn_settings'
    resp, resp_code = get_hash(agent_ip, url, t, waf_id)

    # response
    if resp_code == 200:
        return Response(resp, status=resp_code)
    else:
        return Response(status=resp_code)


@app.route(get_dyn_settings_uri, methods=['POST'])
def get_dyn_settings_f():

    # set URL
    url = request.path

    try:

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.args.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        query = """ SELECT to_jsonb(array_agg(dyn_settings)) FROM dyn_settings WHERE (waf_id = %s AND active IS TRUE) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res:

            # init
            resp = {'status': 'none'}

            # format result
            if resp_in_json:
                resp = json.dumps(resp)
            else:
                resp = str(resp)

            # response
            return Response(resp, status=200)

        else:

            # extract first item from list
            # (only 1 item must be received from the database)
            res = res[0]

            # set status
            res['status'] = 'success'

            # del columns
            del res['active']
            del res['modify']
            del res['waf_id']

            # delete keys with None/empty value
            res = {k: v for k, v in res.items() if v is not None and v != ''}

            # format result
            if resp_in_json:
                resp = json.dumps(res)
            else:
                resp = str(res)

            # response
            return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_dyn_settings_uri, methods=['POST'])
def set_dyn_settings_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in jdata:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = dyn_settings_cols + ['active']

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            ##
            # Create row if it does not exist
            ##

            try:

                # insert the data
                query = """ INSERT INTO dyn_settings (waf_id) VALUES (%s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (waf_id,))

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

            ##
            # Processing request data
            ##

            for k in cols_ext:

                # processing data
                if k in jdata[cmd]:

                    # convert to srt and replace newline/multiple space to space
                    v = re.sub(r'\s+', ' ', str(jdata[cmd][k]))

                    # data preprocessing according to the type of settings
                    if 'nwaf_b64_decode_disable' == k and re.match(r'\sz:$', v):
                        v = v.replace('z:', 'z:URL|ARGS|HEADERS|BODY')

                    # value validations
                    status = dyn_value_validation(agent_ip, k, v)

                    # check status
                    if not status:

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = str('Incorrect value for {}'.format(k))
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            ##
            # Save data to the database
            ##

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                return Response(resp, status=201)

            try:

                try:

                    # prepare SQL data
                    k_list.extend(['modify'])
                    v_list.extend([str(datetime.now())])

                    # update data
                    query = """ UPDATE dyn_settings SET ({}) = %s WHERE (waf_id = %s) """.format(','.join(k_list))
                    cursor.execute(query, (tuple(v_list), waf_id,))

                except Exception as e:

                    # logging
                    log_key = 'x' * 26 + key[-6:]
                    log.info('An error occurred while processing data for key "{}" in dyn_settings: {}'.format(log_key, e))

                    # response
                    close_db_connection(connect, cursor)
                    resp = str({'status': 'fail', 'description': str(e)})
                    return Response(resp, status=201)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

        # processing command
        elif cmd == 'del':

            ##
            # Delete settings from the database
            ##

            try:

                try:

                    query = """ DELETE FROM dyn_settings WHERE (waf_id = %s) """
                    cursor.execute(query, (waf_id,))

                except Exception as e:

                    # logging
                    log_key = 'x' * 26 + key[-6:]
                    log.info('An error occurred while processing data for key "{}" in dyn_settings: {}'.format(log_key, e))

                    # response
                    close_db_connection(connect, cursor)
                    resp = str({'status': 'fail', 'description': str(e)})
                    return Response(resp, status=201)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_dyn_erl_hash_uri, methods=['POST'])
def get_dyn_erl_hash_f():

    # set URL
    url = request.path

    # extract client IP
    agent_ip = get_agent_ip(request)

    # extract lic. key
    key = request.values.get('key')
    if not key:
        log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
        return Response(status=401)

    # WAF ID check
    waf_id = get_waf_id(key, agent_ip)
    if not waf_id:
        return Response(status=401)

    # processing request
    t = 'dyn_erl_ts'
    resp, resp_code = get_hash(agent_ip, url, t, waf_id)

    # response
    if resp_code == 200:
        return Response(resp, status=resp_code)
    else:
        return Response(status=resp_code)


@app.route(get_dyn_erl_uri, methods=['POST'])
def get_dyn_erl_f():

    # set URL
    url = request.path

    try:

        # init
        get_extended = False

        # extract output format
        try:
            args = request.args.get('extended').lower()
            if args == 'yes':
                get_extended = True
        except Exception:
            pass

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.args.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_erl)) from dyn_erl WHERE (waf_id = %s AND active IS TRUE) """
        if get_extended or resp_in_json:
            query = """ SELECT to_jsonb(array_agg(dyn_erl)) from dyn_erl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        ##
        # Response in JSON format
        ##

        # processing result
        if not res or not res[0][0]:

            # init
            resp = {'status': 'none'}

            # format result
            if resp_in_json:
                resp = json.dumps(resp)
            else:
                resp = str(resp)

            # response
            return Response(resp, status=200)

        else:

            # init
            resp = {'status': 'success'}

            # response in JSON
            if resp_in_json:

                # init
                r = []

                # processing columns
                for i in res[0][0]:

                    # add rid key
                    i['rid'] = i['id']

                    # remove unused
                    del i['waf_id']

                    # update the list
                    r.append(i)

                # update response
                resp['erl'] = r

                # convert to JSON
                resp = json.dumps(resp)

            else:

                # init
                resp = str(resp)

                # processing each row
                for row in res[0][0]:
                    try:

                        # delete unused columns
                        for k in row.copy().keys():

                            # if is get_extended
                            if get_extended:
                                if k == 'active':
                                    # don't delete key
                                    continue

                            # other delete conditions
                            if k != 'id' and k not in dyn_erl_cols:
                                del row[k]

                        # update empty columns
                        row = {k: ('' if v is None else v) for k, v in row.items()}

                        ##
                        # update other_headers
                        ##

                        # init
                        other_headers = []

                        # update other_headers
                        if not row['other_headers']:
                            row['other_headers'] = other_headers
                        else:
                            row['other_headers'] = row['other_headers'].split(', ')

                        if not get_extended:

                            # update values for dyn
                            row['id'] = str(row['id'])
                            row['lm'] = str(row['lm']).lower()
                            row['api'] = str(row['api']).lower()
                            row['noban'] = str(row['noban']).lower()
                            row['no_cookie'] = str(row['no_cookie']).lower()

                        # update result
                        resp = resp + '\n' + json.dumps(row)

                    except Exception as e:
                        log.error('An error occurred while processing result from table "dyn_erl": {}'.format(e))

            # response
            return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_dyn_erl_uri, methods=['POST'])
def set_dyn_erl_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error('An error occurred while extracting product key from {} to {}: {}'.format(agent_ip, url, e))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in dyn_wl_erl_cmd:
            if i in jdata:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        ##
        # Processing JSON
        ##

        # init
        t = 'dyn_erl_ts'
        k_list = []
        v_list = []

        # create extended columns
        cols_ext = dyn_erl_cols + ['active']

        # processing command
        if cmd == 'add' or cmd == 'upd':

            # convert all keys to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            ##
            # Checking data compatibility
            ##

            # If no_cookie and cookie in data
            if 'no_cookie' in jdata[cmd] and 'cookie' in jdata[cmd]:

                # Cannot be used no_cookie:true with non-empty cookie
                if jdata[cmd]['no_cookie'] and len(jdata[cmd]['cookie']):
                    err_msg = str('Cannot be used at the same time: "no_cookie" with true and non-empty data in "cookie"')
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

            # 'add' only
            if cmd == 'add':

                # Checking for at least 1 required non-empty key
                if not {k.lower(): v for k, v in jdata[cmd].items() if k in dyn_erl_type_data and v}:
                    close_db_connection(connect, cursor)
                    err_msg = str('At least 1 required key is missing: {}'.format(', '.join(dyn_erl_type_data)))
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

            # processing jdata
            for k in cols_ext:
                try:

                    if k in jdata[cmd]:

                        # convert to srt and replace newline/multiple space to space
                        v = re.sub(r'\s+', ' ', str(jdata[cmd][k]))

                        ##
                        # Value delimiter validation
                        ##

                        # other_headers
                        if k == 'other_headers':

                            # represent values as list
                            ivl = ast.literal_eval(v)

                            # processing list
                            for it1 in ivl:

                                # split value to list
                                it1_l = it1.split()

                                # processing list
                                for it2 in it1_l:

                                    # split to k/v and extract value
                                    ik = it2.split(':')[0].strip()
                                    iv = it2.split(':')[1].strip()

                                    ##
                                    # Processing key
                                    ##

                                    # remove !! from the start before processing
                                    ik = re.sub(r'^!!', '', ik)

                                    # processing key
                                    for pattern in ['!', '&', '|']:
                                        if ik.startswith(pattern) or ik.endswith(pattern):
                                            close_db_connection(connect, cursor)
                                            err_msg = str('The value cannot start/end with the operator "{}" in {}'.format(pattern, ik))
                                            resp = str({'status': 'fail', 'description': err_msg})
                                            return Response(resp, status=201)

                                    ##
                                    # Processing value
                                    ##

                                    # processing value
                                    for pattern in ['&!', '|!']:
                                        if iv.startswith(pattern) or iv.endswith(pattern):
                                            close_db_connection(connect, cursor)
                                            err_msg = str('The value cannot start/end with the operator "{}" in {}'.format(pattern, iv))
                                            resp = str({'status': 'fail', 'description': err_msg})
                                            return Response(resp, status=201)

                                    # processing value
                                    for pattern in ['!', '&', '|']:
                                        if iv.endswith(pattern):
                                            close_db_connection(connect, cursor)
                                            err_msg = str('The value cannot end with the operator "{}" in {}'.format(pattern, iv))
                                            resp = str({'status': 'fail', 'description': err_msg})
                                            return Response(resp, status=201)

                        # other non-boolean keys except other_headers
                        elif k != 'other_headers' and k not in dyn_erl_type_boolean:

                            # init
                            iv = v

                            # processing value
                            for pattern in ['&!', '|!']:
                                if iv.startswith(pattern) or iv.endswith(pattern):
                                    close_db_connection(connect, cursor)
                                    err_msg = str('The value cannot start/end with the operator "{}" in {}'.format(pattern, iv))
                                    resp = str({'status': 'fail', 'description': err_msg})
                                    return Response(resp, status=201)

                            # processing value
                            for pattern in ['!', '&', '|']:
                                if iv.endswith(pattern):
                                    close_db_connection(connect, cursor)
                                    err_msg = str('The value cannot end with the operator "{}" in {}'.format(pattern, iv))
                                    resp = str({'status': 'fail', 'description': err_msg})
                                    return Response(resp, status=201)

                        # value validations
                        status = dyn_erl_value_validation(agent_ip, k, v)

                        # check status
                        if not status:

                            # response
                            close_db_connection(connect, cursor)
                            err_msg = str('Incorrect value for {}'.format(k))
                            resp = str({'status': 'fail', 'description': err_msg})
                            return Response(resp, status=201)

                        # replace quotes in list
                        if isinstance(jdata[cmd][k], list):
                            v = ', '.join(jdata[cmd][k])

                        # update key/value lists
                        k_list.append(k)
                        v_list.append(v)

                except Exception as e:
                    log.error('An error occurred in {} while processing data: {}'.format(url, e))
                    close_db_connection(connect, cursor)
                    return Response(status=503)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                return Response(resp, status=201)

        ##
        # Processing the command
        ##

        if cmd == 'add':
            try:

                # prepare SQL data
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_erl ({}) VALUES %s RETURNING ID """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list),))
                res = cursor.fetchone()

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # set response
                if res:
                    # if ID was returned
                    resp = str({'status': 'success', 'rid': res[0]})
                else:
                    # if ID was not returned
                    resp = str({'status': 'fail', 'description': 'ID was not obtained while saving the rule'})

                # close database connection
                close_db_connection(connect, cursor)

                # response
                return Response(resp, status=201)

            except Exception as e:

                # set error message
                err_msg = str(e)

                # processing err_msg
                if 'duplicate key value' in err_msg:
                    err_msg = 'Duplicate key value'

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

        elif cmd == 'upd':
            try:

                """
                Rules are displayed and could be edited for other members with the same WAF ID.
                """

                ##
                # Processing request data
                ##

                # extract value
                param = 'rid'
                if param in jdata[cmd]:
                    try:
                        rid = int(jdata[cmd][param])
                    except Exception:
                        close_db_connection(connect, cursor)
                        resp = str({'status': 'fail', 'description': 'Parameter "rid" is not an integer'})
                        return Response(resp, status=201)

                # if no ID in the request
                else:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'No ID received'
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

                ##
                # Search for rule ID (to update the hash in the future)
                ##

                query = """ SELECT id from dyn_erl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

                ##
                # Checking data compatibility:
                # at least 1 key from dyn_erl_type_data will be in the rule
                ##

                # if at least 1 key from dyn_erl_type_data in the jdata[cmd]
                if [i for i in dyn_erl_type_data if i in jdata[cmd].keys()]:

                    query = """ SELECT {} from dyn_erl WHERE (waf_id = %s AND id = %s) """.format(','.join(dyn_erl_type_data))
                    cursor.execute(query, (waf_id, rid,))
                    res = cursor.fetchone()

                    # check result
                    if res:

                        # init
                        res_dict = {}

                        # create PG response dict
                        for i, k in enumerate(dyn_erl_type_data):
                            res_dict[k] = res[i]

                        # remove unused columns
                        for k in res_dict.copy().keys():
                            if k not in dyn_erl_type_data:
                                res_dict.pop(k)

                        # merge result from the database with data from the request
                        for k in res_dict.copy().keys():
                            if k in jdata[cmd]:
                                res_dict[k] = jdata[cmd][k]

                        # delete keys with None/empty value
                        res_dict = {k: v for k, v in res_dict.items() if v is not None and v != '' and v != '[]'}

                        # If res_dict do not contain at least 1 non-empty key
                        if not len(res_dict):
                            close_db_connection(connect, cursor)
                            err_msg = str(
                                'After updating, the rule must contain at least 1 required key: {}'
                                .format(', '.join(dyn_erl_type_data))
                            )
                            resp = str({'status': 'fail', 'description': err_msg})
                            return Response(resp, status=201)

                    else:

                        """
                        Row is not found, return response
                        """

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'ID {} not found in database'.format(rid)
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                ##
                # Update data
                ##

                # prepare SQL data
                k_list.extend(['modify'])
                v_list.extend([str(datetime.now())])

                # insert the data
                query = """ UPDATE dyn_erl SET ({}) = %s WHERE (waf_id = %s AND id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

        elif cmd == 'del':
            try:

                # convert all keys to lowercase
                jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

                # extract value
                param = 'rid'
                if param in jdata[cmd]:
                    try:
                        rid = int(jdata[cmd][param])
                    except Exception:
                        close_db_connection(connect, cursor)
                        resp = str({'status': 'fail', 'description': 'Parameter "rid" is not an integer'})
                        return Response(resp, status=201)

                # if no ID in the request
                else:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'No ID received'
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

                try:

                    query = """ SELECT id from dyn_erl WHERE (waf_id = %s AND id = %s) """
                    cursor.execute(query, (waf_id, rid,))
                    res = cursor.fetchone()

                    # check result
                    if not res:

                        """
                        Row is not found, return response
                        """

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = 'ID {} not found in database'.format(rid)
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # search data
                    query = """ DELETE from dyn_erl WHERE (waf_id = %s AND id = %s) """
                    cursor.execute(query, (waf_id, rid,))

                    # update ts
                    set_hash(agent_ip, url, t, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    resp = str({'status': 'success'})
                    return Response(resp, status=201)

                except Exception as e:

                    # response
                    close_db_connection(connect, cursor)
                    resp = str({'status': 'fail', 'description': str(e)})
                    return Response(resp, status=201)

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_dyn_wl_hash_uri, methods=['POST'])
def get_dyn_wl_hash_f():

    # set URL
    url = request.path

    # extract client IP
    agent_ip = get_agent_ip(request)

    # extract lic. key
    key = request.values.get('key')
    if not key:
        log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
        return Response(status=401)

    # WAF ID check
    waf_id = get_waf_id(key, agent_ip)
    if not waf_id:
        return Response(status=401)

    # processing request
    t = 'dyn_wl_ts'
    resp, resp_code = get_hash(agent_ip, url, t, waf_id)

    # response
    if resp_code == 200:
        return Response(resp, status=resp_code)
    else:
        return Response(status=resp_code)


@app.route(get_dyn_wl_uri, methods=['POST'])
def get_dyn_wl_f():

    # set URL
    url = request.path

    try:

        # init
        get_extended = False

        # extract output format
        try:
            args = request.args.get('extended').lower()
            if args == 'yes':
                get_extended = True
        except Exception:
            pass

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.args.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        # search in the database
        query = """ SELECT to_jsonb(array_agg(dyn_wl)) FROM dyn_wl WHERE (waf_id = %s AND active IS TRUE) """
        if get_extended or resp_in_json:
            query = """ SELECT to_jsonb(array_agg(dyn_wl)) FROM dyn_wl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res or not res[0][0]:

            # init
            resp = {'status': 'none'}

            # format result
            if resp_in_json:
                resp = json.dumps(resp)
            else:
                resp = str(resp)

            # response
            return Response(resp, status=200)

        else:

            # init
            resp = {'status': 'success'}

            # response in JSON
            if resp_in_json:

                # init
                r = []

                # processing columns
                for i in res[0][0]:

                    # rename column
                    i['rid'] = i['id']
                    del i['id']

                    # remove unused
                    del i['waf_id']

                    # update the list
                    r.append(i)

                # update response
                resp['wl'] = r

                # convert to JSON
                resp = json.dumps(resp)

            else:

                # init
                resp = str(resp)

                # processing each row
                for row in res[0][0]:
                    try:

                        # init
                        r = 'WL'

                        # ID processing
                        r = r + ' ' + 'ID:' + str(row['rl_id'])

                        # Domain processing
                        if row['domain']:
                            r = r + ' domain=' + str(row['domain'])

                        ##
                        # Z processing
                        ##

                        if not row['mz']:
                            row['mz'] = ''
                        if not row['extension']:
                            row['extension'] = ''

                        mz_ext = list(filter(None, (row['mz'].split('|') + row['extension'].split('|'))))
                        r = r + ' "' + 'Z:' + '|'.join(mz_ext) + '"'

                        # update result
                        if r:

                            if get_extended:
                                r = 'RID: ' + str(row['id']) + ' Active: ' + str(row['active']) + ' ' + r

                            resp = resp + '\n' + r

                    except Exception as e:
                        log.error(
                            'An error occurred while processing result from table "dyn_wl": {}'
                            .format(e)
                        )

            # response
            return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_dyn_wl_uri, methods=['POST'])
def set_dyn_wl_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in dyn_wl_erl_cmd:
            if i in jdata:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        ##
        # Processing JSON
        ##

        # init
        t = 'dyn_wl_ts'

        # create extended columns
        cols_ext = dyn_wl_cols + ['active']

        # processing command
        if cmd == 'add':

            # init
            k_list = []
            v_list = []

            # convert all keys to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            # extract rl_id
            param = 'rl_id'
            if param not in jdata[cmd]:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'No rl_id was received'})
                return Response(resp, status=201)

            # processing jdata
            for k in cols_ext:

                if k in jdata[cmd]:

                    # convert to srt and replace newline/multiple space to space
                    v = re.sub(r'\s+', ' ', str(jdata[cmd][k]))

                    # value validations
                    status = dyn_wl_value_validation(agent_ip, k, v)

                    # check status
                    if not status:

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = str('Incorrect value for {}'.format(k))
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                return Response(resp, status=201)

            ##
            # Create row
            ##

            try:

                # init
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_wl ({}) VALUES %s RETURNING ID """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list),))
                res = cursor.fetchone()

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # set response
                if res:
                    # if ID was returned
                    resp = str({'status': 'success', 'rid': res[0]})
                else:
                    # if ID was not returned
                    resp = str({'status': 'fail', 'description': 'ID was not obtained while saving the rule'})

                # close database connection
                close_db_connection(connect, cursor)

                # response
                return Response(resp, status=201)

            except Exception as e:

                # set error message
                err_msg = str(e)

                # processing err_msg
                if 'duplicate key value' in err_msg:
                    err_msg = 'Duplicate key value'

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

        if cmd == 'upd':

            # init
            k_list = []
            v_list = []

            # convert all keys/values to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            # extract value
            param = 'rid'
            if param in jdata[cmd]:
                try:
                    rid = int(jdata[cmd][param])
                except Exception:
                    resp = str({'status': 'fail', 'description': 'Parameter "rid" is not an integer'})
                    return Response(resp, status=201)

            # if no ID in the request
            else:

                # response
                close_db_connection(connect, cursor)
                err_msg = 'No ID received'
                resp = str({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

            ##
            # Extract RL ID
            ##

            query = """ SELECT id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
            cursor.execute(query, (waf_id, rid,))
            res = cursor.fetchone()

            # set type
            if not res:

                # response
                close_db_connection(connect, cursor)
                err_msg = 'ID {} not found in database'.format(rid)
                resp = str({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

            # processing jdata
            for k in cols_ext:

                if k in jdata[cmd]:

                    # convert to srt and replace newline/multiple space to space
                    v = re.sub(r'\s+', ' ', str(jdata[cmd][k]))

                    # value validations
                    status = dyn_wl_value_validation(agent_ip, k, v)

                    # check status
                    if not status:

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = str('Incorrect value for {}'.format(k))
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                return Response(resp, status=201)

            ##
            # Update row
            ##

            try:

                ##
                # Processing request data
                ##

                query = """ SELECT waf_id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

                # init
                k_list.append('modify')
                v_list.append(str(datetime.now()))

                # update data
                query = """ UPDATE dyn_wl SET ({}) = %s WHERE (waf_id = %s AND id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

        # processing command
        elif cmd == 'del':

            # convert all keys to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            # extract value
            param = 'rid'
            if param in jdata[cmd]:
                try:
                    rid = int(jdata[cmd][param])
                except Exception:
                    close_db_connection(connect, cursor)
                    resp = str({'status': 'fail', 'description': 'Parameter "rid" is not an integer'})
                    return Response(resp, status=201)

            # if no ID in the request
            else:

                # response
                close_db_connection(connect, cursor)
                err_msg = 'No ID received'
                resp = str({'status': 'fail', 'description': err_msg})
                return Response(resp, status=201)

            try:

                # search data
                query = """ SELECT waf_id FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))
                res = cursor.fetchone()

                # check result
                if not res:

                    """
                    Row is not found, return response
                    """

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = 'ID {} not found in database'.format(rid)
                    resp = str({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

                query = """ DELETE FROM dyn_wl WHERE (waf_id = %s AND id = %s) """
                cursor.execute(query, (waf_id, rid,))

                # update ts
                set_hash(agent_ip, url, t, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_dyn_bl_hash_uri, methods=['POST'])
def get_dyn_bl_hash_f():

    # set URL
    url = request.path

    # extract client IP
    agent_ip = get_agent_ip(request)

    # extract lic. key
    key = request.values.get('key')
    if not key:
        log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
        return Response(status=401)

    # WAF ID check
    waf_id = get_waf_id(key, agent_ip)
    if not waf_id:
        return Response(status=401)

    # processing request
    t = 'dyn_bl'
    resp, resp_code = get_hash(agent_ip, url, t, waf_id)

    # response
    if resp_code == 200:
        return Response(resp, status=resp_code)
    else:
        return Response(status=resp_code)


@app.route(get_dyn_bl_uri, methods=['POST'])
def get_dyn_bl_f():

    # set URL
    url = request.path

    try:

        # init
        resp_in_json = False

        # extract output format
        try:
            args = request.args.get('format').lower()
            if args == 'json':
                resp_in_json = True
        except Exception:
            pass

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        # search in the database
        query = """ SELECT active, bl FROM dyn_bl WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()

        # close database connection
        close_db_connection(connect, cursor)

        ##
        # Response in JSON format
        ##

        # processing result
        if not res:

            # init
            resp = {'status': 'none'}

            # format result
            if resp_in_json:
                resp = json.dumps(resp)
            else:
                resp = str(resp)

            # response
            return Response(resp, status=200)

        else:

            # init
            status = res[0]
            bl = ast.literal_eval(res[1])

            if resp_in_json:

                # convert to JSON
                resp = json.dumps(
                    {'status': 'success', 'active': status, 'bl': bl}
                )

            else:

                # init
                resp = str({'status': 'success'})

                # processing BL
                if len(bl):
                    resp = resp + '\n' + '\n'.join(bl)

            # response
            return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_dyn_bl_uri, methods=['POST'])
def set_dyn_bl_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Check allowed command
        ##

        # init
        cmd = ''

        # search for allowed command in JSON
        for i in settings_cmd:
            if i in jdata:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        ##
        # Processing JSON
        ##

        # create extended columns
        cols_ext = dyn_bl_cols + ['active']

        # processing command
        if cmd == 'set':

            # init
            k_list = []
            v_list = []

            # convert all keys/values to lowercase
            jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

            # processing jdata
            for k in cols_ext:

                if k in jdata[cmd]:

                    # convert to srt and replace newline/multiple space to space
                    v = re.sub(r'\s+', ' ', str(jdata[cmd][k]))

                    # value validations
                    status = dyn_bl_value_validation(agent_ip, k, v)

                    # check status
                    if not status:

                        # response
                        close_db_connection(connect, cursor)
                        err_msg = str('Incorrect value for {}'.format(k))
                        resp = str({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=201)

                    # sort unique and replace quotes in list
                    if isinstance(jdata[cmd][k], list):
                        v = list(set(jdata[cmd][k]))
                        v = str(v).replace("'", '"')

                    # update key/value lists
                    k_list.append(k)
                    v_list.append(v)

            # check JSON parameters
            if not k_list:
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                return Response(resp, status=201)

            ##
            # Update row
            ##

            try:

                ##
                # Processing request data
                ##

                # prepare SQL data
                k_list.extend(['waf_id', 'modify'])
                v_list.extend([waf_id, str(datetime.now())])

                # insert the data
                query = """ INSERT INTO dyn_bl ({}) VALUES %s ON CONFLICT (waf_id) DO UPDATE SET ({}) = %s """.format(
                    ','.join(k_list), ','.join(k_list)
                )
                cursor.execute(query, (tuple(v_list), tuple(v_list),))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

        # processing command
        elif cmd == 'del':
            try:

                # delete data
                query = """ DELETE FROM dyn_bl WHERE (waf_id = %s) """
                cursor.execute(query, (waf_id,))

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = str({'status': 'success'})
                return Response(resp, status=201)

            except Exception as e:

                # response
                close_db_connection(connect, cursor)
                resp = str({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


##
# Nginx Settings
##

@app.route(get_nginx_status_uri, methods=['GET'])
def get_nginx_status_f():

    # set URL
    url = request.path

    try:

        # response
        resp = json.dumps({'is_active': True})
        return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_nginx_global_hash_uri, methods=['POST'])
@app.route(get_nginx_vhost_hash_uri, methods=['POST'])
@app.route(get_nginx_upstream_hash_uri, methods=['POST'])
@app.route(get_nginx_certificate_hash_uri, methods=['POST'])
@app.route(get_nginx_dhparam_hash_uri, methods=['POST'])
def get_nginx_hash_f():

    # set URL
    url = request.path

    try:

        # set table
        if url == get_nginx_global_hash_uri:
            table = 'nginx_' + 'global'
        elif url == get_nginx_vhost_hash_uri:
            table = 'nginx_' + 'vhost'
        elif url == get_nginx_upstream_hash_uri:
            table = 'nginx_' + 'upstream'
        elif url == get_nginx_certificate_hash_uri:
            table = 'nginx_' + 'certificate'
        elif url == get_nginx_dhparam_hash_uri:
            table = 'nginx_' + 'dhparam'
        else:
            err_msg = 'An unknown request path {}'.format(request.path)
            log.error(err_msg)
            resp = json.dumps({'status': 'fail', 'description': err_msg})
            return Response(resp, status=200)

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            return Response(status=401)

        # Extract the hash
        res = get_nginx_hash(table, waf_id)

        # processing result
        if not res:
            resp = {'status': 'none'}
        else:
            resp = {'status': 'success', 'hash': res}

        # response
        resp = json.dumps(resp)
        return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_nginx_global_uri, methods=['POST'])
@app.route(get_nginx_vhost_uri, methods=['POST'])
@app.route(get_nginx_upstream_uri, methods=['POST'])
@app.route(get_nginx_dhparam_uri, methods=['POST'])
@app.route(get_nginx_certificate_uri, methods=['POST'])
@app.route(get_nginx_dhparam_secure_uri, methods=['POST'])
@app.route(get_nginx_certificate_secure_uri, methods=['POST'])
def get_nginx_f():

    # set URL
    url = request.path

    try:

        # set type
        if url == get_nginx_global_uri:
            nginx_type = 'global'
        elif url == get_nginx_vhost_uri:
            nginx_type = 'vhost'
        elif url == get_nginx_upstream_uri:
            nginx_type = 'upstream'
        elif url in [get_nginx_dhparam_uri, get_nginx_dhparam_secure_uri]:
            nginx_type = 'dhparam'
        elif url in [get_nginx_certificate_uri, get_nginx_certificate_secure_uri]:
            nginx_type = 'certificate'
        else:
            err_msg = 'An unknown request path {}'.format(request.path)
            log.error(err_msg)
            resp = json.dumps({'status': 'fail', 'description': err_msg})
            return Response(resp, status=200)

        # set table
        table = 'nginx_' + nginx_type

        # set response format
        resp_in_nginx_format = False
        if url in get_nginx_with_format:
            try:
                args = request.args.get('format').lower()
                if args == 'nginx':
                    resp_in_nginx_format = True
            except Exception:
                pass

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract lic. key
        key = request.values.get('key')
        if not key:
            log.error('The product key was not received from client {} in {}'.format(url, agent_ip))
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        # init ORDER BY
        if url != get_nginx_global_uri:
            order_by = 'ORDER BY name'
        else:
            order_by = ''

        # set the query
        if url in get_nginx_with_active and resp_in_nginx_format:
            # if table with 'active' column and response in Nginx format
            query = """ SELECT to_jsonb(array_agg({} {})) FROM {} WHERE (waf_id = %s AND active IS TRUE) """.format(table, order_by, table)
        else:
            query = """ SELECT to_jsonb(array_agg({} {})) FROM {} WHERE (waf_id = %s) """.format(table, order_by, table)

        # get data from the database
        cursor.execute(query, (waf_id,))
        res = cursor.fetchall()

        # close database connection
        close_db_connection(connect, cursor)

        # processing result
        if not res or not res[0][0]:

            if resp_in_nginx_format:
                resp = ''
            else:
                resp = {'status': 'none'}
                resp = json.dumps(resp)

            # response
            return Response(resp, status=200)

        else:
            try:

                # init
                resp = ''
                data_list = res[0][0]

                # Nginx format
                if resp_in_nginx_format:

                    # data processing
                    for idx, data in enumerate(data_list.copy()):

                        # vhost processing
                        if nginx_type == 'vhost' and 'name' in data:
                            del data['name']

                        # remove unused columns
                        for k in ['active', 'waf_id']:
                            if k in data:
                                del data[k]

                        # remove empty/None values
                        data = {k: v for k, v in data.items() if (v != '' and v is not None)}

                        # update data list
                        data_list[idx] = data

                    # convert data to Nginx format
                    for data in data_list:
                        resp = resp + nginx_format_processing(nginx_type, data)

                # JSON format
                else:

                    # init
                    res = {}

                    # add status
                    for idx, data in enumerate(data_list.copy()):

                        # remove unused columns
                        if 'waf_id' in data:
                            del data['waf_id']

                        # data processing
                        for k, v in data.items():

                            # dhparam/private key masking for non-secure URL
                            if url in [get_nginx_dhparam_uri, get_nginx_certificate_uri]:
                                if k in ['dhparam', 'certificate_key']:
                                    try:
                                        data[k] = base64.b64encode(
                                            base64.b64decode(v.encode())[:40]
                                            + b'*****'
                                            + base64.b64decode(v.encode())[-40:]
                                        ).decode()
                                    except Exception as e:
                                        log.error('An error occurred while masking the {} in {}: {}'.format(k, url, e))
                                        data[k] = ''

                            # list / list of dictionary  normalization
                            if k in nginx_type_list + nginx_type_dict + nginx_type_lod:
                                if v is not None:
                                    data[k] = ast.literal_eval(v)

                        # update data list
                        data_list[idx] = data

                    # update result
                    res['status'] = 'success'
                    res['configuration'] = data_list

                    # response
                    resp = json.dumps(res)

                # response
                return Response(resp, status=200)

            except Exception as e:
                log.error('An error occurred while processing result from table "{}": {}'.format(table, e))

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_nginx_global_uri, methods=['POST'])
@app.route(set_nginx_vhost_uri, methods=['POST'])
@app.route(set_nginx_upstream_uri, methods=['POST'])
@app.route(set_nginx_dhparam_uri, methods=['POST'])
@app.route(set_nginx_certificate_uri, methods=['POST'])
def set_nginx_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # set type
        if url == set_nginx_global_uri:
            nginx_type = 'global'
            cols = nginx_global_cols
        elif url == set_nginx_vhost_uri:
            nginx_type = 'vhost'
            cols = nginx_vhost_cols
        elif url == set_nginx_upstream_uri:
            nginx_type = 'upstream'
            cols = nginx_upstream_cols
        elif url == set_nginx_certificate_uri:
            nginx_type = 'certificate'
            cols = nginx_certificate_cols
        elif url == set_nginx_dhparam_uri:
            nginx_type = 'dhparam'
            cols = nginx_dhparam_cols
        else:
            err_msg = 'An unknown request path {}'.format(request.path)
            log.error(err_msg)
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': err_msg})
            return Response(resp, status=200)

        # set table
        table = 'nginx_' + nginx_type

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # WAF ID check
        waf_id = get_waf_id(key, agent_ip)
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Check allowed command
        ##

        # init
        cmd = ''
        name = ''

        # search for allowed command in JSON
        for i in nginx_cmd:
            if i in jdata:
                cmd = i.lower()
                break

        # processing result
        if not cmd:

            # logging
            log.error('Required key missing in JSON from {} to {}'.format(agent_ip, url))

            # response
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=200)

        ##
        # Processing JSON
        ##

        # convert all keys to lowercase
        jdata[cmd] = {k.lower(): v for k, v in jdata[cmd].items()}

        # required key check
        if table in ['nginx_vhost', 'nginx_upstream', 'nginx_certificate', 'nginx_dhparam']:
            req_k = 'name'
            if req_k not in jdata[cmd]:
                err_msg = 'The required key "{}" is missing'.format(req_k)
                log.error(err_msg)
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': err_msg})
                return Response(resp, status=200)
            else:
                name = jdata[cmd]['name']

        # processing command
        if cmd == 'set':
            try:

                # init
                data = {}

                # processing jdata
                for k in cols:

                    if k in jdata[cmd]:

                        # init
                        v = jdata[cmd][k]

                        # value validations
                        r = nginx_validation(waf_id, nginx_type, jdata[cmd], k, v)
                        if not r[0]:
                            close_db_connection(connect, cursor)
                            resp = json.dumps({'status': 'fail', 'description': r[1]})
                            return Response(resp, status=200)

                        # convert list/dictionary to string
                        if isinstance(v, list) or isinstance(v, dict):
                            v = str(v)

                        # update value lists
                        data[k] = v

                # check JSON parameters
                if not data:
                    close_db_connection(connect, cursor)
                    resp = json.dumps({'status': 'fail', 'description': 'Parameters are missing in JSON'})
                    return Response(resp, status=200)
                else:
                    # add WAF ID
                    data['waf_id'] = waf_id

                # init
                k = list(data.keys())
                v = list(data.values())

                # set conflict columns list
                key_columns = ['waf_id'] if table == 'nginx_global' else ['waf_id', 'name']

                # insert the data
                query = """ INSERT INTO {} ({}) VALUES %s ON CONFLICT ({}) DO UPDATE SET ({}) = %s """.format(
                    table, ','.join(k), ','.join(key_columns), ','.join(k)
                )
                cursor.execute(query, (tuple(v), tuple(v),))

                # update the hash
                set_nginx_hash(table, waf_id)

                # close database connection
                close_db_connection(connect, cursor)

                # response
                resp = json.dumps({'status': 'success'})
                return Response(resp, status=200)

            except Exception as e:
                log.error('An error occurred while set the Nginx settings: {}'.format(e))
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': str(e)})
                return Response(resp, status=200)

        elif cmd == 'del':
            try:

                if table not in ['nginx_vhost', 'nginx_upstream', 'nginx_certificate', 'nginx_dhparam']:
                    err_msg = 'The command "{}" is not applicable to {}'.format(cmd, url)
                    log.error(err_msg)
                    close_db_connection(connect, cursor)
                    resp = json.dumps({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=200)

                # checking that the entity is not used
                if nginx_type == 'vhost':

                    # delete vhost
                    query = """ DELETE FROM {} WHERE (waf_id = %s AND name = %s) """.format(table)
                    cursor.execute(query, (waf_id, name,))

                    # update the hash
                    set_nginx_hash(table, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    resp = json.dumps({'status': 'success'})
                    return Response(resp, status=200)

                elif nginx_type in ['certificate', 'dhparam']:

                    # search entity in nginx_vhost
                    column = 'certificate_name' if nginx_type == 'certificate' else 'dhparam_name'
                    query = """ SELECT COUNT(*) from nginx_vhost WHERE (waf_id = %s AND {} = %s) """.format(column)
                    cursor.execute(query, (waf_id, name,))
                    res = cursor.fetchone()

                    # processing result
                    if not res or res[0] < 1:
                        query = """ DELETE FROM {} WHERE (waf_id = %s AND name = %s) """.format(table)
                        cursor.execute(query, (waf_id, name,))
                    else:
                        err_msg = 'The {} "{}" in use'.format(nginx_type, name)
                        resp = json.dumps({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=200)

                    # update the hash
                    set_nginx_hash(table, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    resp = json.dumps({'status': 'success'})
                    return Response(resp, status=200)

                elif nginx_type == 'upstream':

                    # search upstream name in the database
                    query = """ SELECT error_page, locations FROM nginx_vhost WHERE (waf_id = %s) """
                    cursor.execute(query, (waf_id,))
                    res = cursor.fetchall()

                    # processing result
                    for row in res:

                        # skip empty
                        if not row:
                            continue

                        # search the upstream in error_page
                        if row[0]:

                            # processing the values
                            for v in ast.literal_eval(row[0]).values():

                                # skip empty value
                                if not v:
                                    continue

                                # strip the value
                                v = v.strip()

                                # if value contains the upstream
                                if v.endswith('@' + name):
                                    close_db_connection(connect, cursor)
                                    err_msg = 'Upstream "{}" is in use'.format(name)
                                    resp = json.dumps({'status': 'fail', 'description': err_msg})
                                    return Response(resp, status=200)

                        # search the upstream in locations
                        if row[1]:

                            # processing for each location
                            for location in ast.literal_eval(row[1]):

                                # skip empty location
                                if not location:
                                    continue

                                # search upstream in proxy_pass
                                k = 'proxy_pass'
                                if k in location and location[k].strip().endswith('://' + name):
                                    close_db_connection(connect, cursor)
                                    err_msg = 'Upstream "{}" is in use'.format(name)
                                    resp = json.dumps({'status': 'fail', 'description': err_msg})
                                    return Response(resp, status=200)

                                # search upstream in memcached_pass
                                k = 'memcached_pass'
                                if k in location and location[k].strip() == name:
                                    close_db_connection(connect, cursor)
                                    err_msg = 'Upstream "{}" is in use'.format(name)
                                    resp = json.dumps({'status': 'fail', 'description': err_msg})
                                    return Response(resp, status=200)

                                # search the upstream in error_page
                                k = 'error_page'
                                if k in location:

                                    # processing the values
                                    for v in location[k].values():

                                        # skip empty value
                                        if not v:
                                            continue

                                        # strip the value
                                        v = v.strip()

                                        # if value contains the upstream
                                        if v.endswith('@' + name):
                                            close_db_connection(connect, cursor)
                                            err_msg = 'Upstream "{}" is in use'.format(name)
                                            resp = json.dumps({'status': 'fail', 'description': err_msg})
                                            return Response(resp, status=200)

                    # delete the entity
                    query = """ DELETE FROM {} WHERE (waf_id = %s AND name = %s) """.format(table)
                    cursor.execute(query, (waf_id, name,))

                    # update the hash
                    set_nginx_hash(table, waf_id)

                    # close database connection
                    close_db_connection(connect, cursor)

                    # response
                    resp = json.dumps({'status': 'success'})
                    return Response(resp, status=200)

            except Exception as e:
                log.error('An error occurred while deleting "{}" entity: {}'.format(nginx_type, e))
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': str(e)})
                return Response(resp, status=200)

        else:
            close_db_connection(connect, cursor)
            err_msg = 'Invalid command received: {}'.format(cmd)
            resp = json.dumps({'status': 'fail', 'description': err_msg})
            return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


##
# Event collector
##

@app.route(get_event_status_uri, methods=['GET'])
def get_event_status_f():

    # set URL
    url = request.path

    try:

        # response
        resp = json.dumps({'is_active': True})
        return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(get_event_uri, methods=['POST'])
def get_event_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id, models_num
        waf_id, _, _, _ = get_auth(key, agent_ip)

        # WAF ID check
        if not waf_id:
            return Response(status=401)

        ##
        # Receive data from the database
        ##

        # init
        resp = {'status': 'success', 'count': 0, 'waf_id': waf_id, 'events': []}

        try:

            # init
            offset = 0
            limit = None
            limit_max = 3000

            # extract the limit
            try:
                k = 'limit'
                if k in jdata:
                    limit = int(jdata[k])
                    if limit > limit_max:
                        limit = limit_max
            except Exception as e:
                log.error(
                    'An error occurred while extracting the limit from {} to {}: {}'
                    .format(agent_ip, url, e)
                )
                close_db_connection(connect, cursor)
                return Response(status=400)

            # extract the offset
            if limit is not None:
                try:
                    k = 'offset'
                    if k in jdata:
                        offset = int(jdata[k]) * limit
                except Exception as e:
                    log.error(
                        'An error occurred while extracting the limit from {} to {}: {}'
                        .format(agent_ip, url, e)
                    )
                    close_db_connection(connect, cursor)
                    return Response(status=400)

            # other parameters processing
            for param in ['starttime', 'endtime', 'ip', 'uuid', 'agent', 'level', 'category']:
                try:

                    # int validation and normalization
                    if param in ['starttime', 'endtime']:
                        if param in jdata and not isinstance(jdata[param], int):
                            return Response(status=400)
                        else:
                            if param not in jdata or jdata[param] is None:
                                jdata[param] = 0 if param == 'starttime' else 9999999999

                    # list validation and normalization
                    if param in ['ip', 'uuid', 'agent', 'level', 'category']:
                        if param in jdata and not isinstance(jdata[param], list):
                            return Response(status=400)
                        else:
                            if param not in jdata or not jdata[param]:
                                jdata[param] = '%%'
                            else:
                                jdata[param] = '|'.join([str(x) for x in jdata[param]])

                except Exception as e:
                    log.error(
                        'An error occurred while extracting data from {} to {}: {}'
                        .format(agent_ip, url, e)
                    )
                    close_db_connection(connect, cursor)
                    return Response(status=400)

            # get the count of all rows
            query = """
                SELECT
                    count(*)
                FROM
                    events
                WHERE (
                    waf_id = %s
                        AND
                    timestamp >= %s
                        AND
                    timestamp <= %s
                        AND
                    ip SIMILAR TO %s
                        AND
                    uuid SIMILAR TO %s
                        AND
                    agent SIMILAR TO %s
                        AND
                    level SIMILAR TO %s
                        AND
                    category SIMILAR TO %s
                )
            """
            cursor.execute(query, (
                waf_id,
                int(jdata['starttime']),
                int(jdata['endtime']),
                jdata['ip'],
                jdata['uuid'],
                jdata['agent'],
                jdata['level'],
                jdata['category'],
            ))
            res = cursor.fetchone()

            # count processing
            if res:

                # update the response data
                resp['count'] = res[0]
                resp['events'] = []

                # search the data
                query = """
                    SELECT
                        id, ip, uuid, agent, level, category, timestamp, convert_from(description, 'utf-8')
                    FROM
                        events
                    WHERE (
                        waf_id = %s
                            AND
                        timestamp >= %s
                            AND
                        timestamp <= %s
                            AND
                        ip SIMILAR TO %s
                            AND
                        uuid SIMILAR TO %s
                            AND
                        agent SIMILAR TO %s
                            AND
                        level SIMILAR TO %s
                            AND
                        category SIMILAR TO %s
                    )
                    ORDER BY id DESC
                    LIMIT %s
                    OFFSET %s
                """
                cursor.execute(query, (
                    waf_id,
                    int(jdata['starttime']),
                    int(jdata['endtime']),
                    jdata['ip'],
                    jdata['uuid'],
                    jdata['agent'],
                    jdata['level'],
                    jdata['category'],
                    limit,
                    offset,
                ))
                res = cursor.fetchall()

                # processing result
                if res:

                    # processing the result
                    for item in tuple(res):

                        # item processing
                        r = {
                            'id': item[0],
                            'ip': item[1],
                            'uuid': item[2],
                            'agent': item[3],
                            'level': item[4],
                            'category': item[5],
                            'timestamp': item[6],
                            'description': item[7]
                        }

                        # update the response data
                        resp['events'].append(r)

                        # del var
                        del r

                # del vars
                del res

        except Exception as e:
            log.error(
                'An error occurred while processing data from the database in {}: {}'
                .format(url, e)
            )

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(json.dumps(resp), status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_event_uri, methods=['POST'])
def set_event_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # extract events
        try:
            events = jdata['events']
        except Exception as e:
            log.error(
                'An error occurred while extracting the events from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=400)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # WAF ID check
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Processing JSON
        ##

        # init
        cols = set_event_cols

        # processing each item in the list
        for event in events:
            try:

                # convert all keys to lowercase
                event = {k.lower(): v for k, v in event.items()}

                # required key check
                for req_k in cols:
                    if req_k not in event.keys():
                        err_msg = 'The required key "{}" is missing'.format(req_k)
                        log.error(err_msg)
                        resp = json.dumps({'status': 'fail', 'description': err_msg})
                        return Response(resp, status=200)

                # processing item
                for k in cols:

                    # init
                    v = event[k]

                    # value validations
                    r = event_validation(k, v)
                    if not r[0]:
                        resp = json.dumps({'status': 'fail', 'description': r[1]})
                        return Response(resp, status=200)

                # add WAF ID
                event['waf_id'] = waf_id

                # decode description
                try:
                    event['description'] = base64.b64decode(event['description']).decode('UTF-8', 'ignore').encode('UTF-8', 'ignore')
                except Exception as e:
                    err = 'An error occurred while decoding the description: {}'.format(e)
                    resp = json.dumps({'status': 'fail', 'description': err})
                    return Response(resp, status=200)

                # init
                k = list(event.keys())
                v = list(event.values())

                # insert the data
                query = """ INSERT INTO events ({}) VALUES %s """.format(','.join(k))
                cursor.execute(query, (tuple(v),))

            except Exception as e:
                log.error('An error occurred while processing the event: {} ({})'.format(e, event))
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': str(e)})
                return Response(resp, status=200)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        resp = json.dumps({'status': 'success'})
        return Response(resp, status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(del_event_uri, methods=['POST'])
def del_event_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = str({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # processing the parameters
        for param in ['timestamp', 'id', 'ip', 'uuid', 'agent', 'level', 'category']:
            try:

                # int validation and normalization
                if param in ['timestamp']:
                    if param in jdata and not isinstance(jdata[param], int):
                        return Response(status=400)
                    else:
                        if param not in jdata or jdata[param] is None:
                            jdata[param] = 0 if param == 'starttime' else 9999999999

                # list validation and normalization
                if param in ['id', 'ip', 'uuid', 'agent', 'level', 'category']:
                    if param in jdata and not isinstance(jdata[param], list):
                        return Response(status=400)
                    else:
                        if param not in jdata or not jdata[param]:
                            jdata[param] = '%%'
                        else:
                            jdata[param] = '|'.join([str(x) for x in jdata[param]])

            except Exception as e:
                log.error(
                    'An error occurred while extracting data from {} to {}: {}'
                    .format(agent_ip, url, e)
                )
                close_db_connection(connect, cursor)
                return Response(status=400)

        # get waf_id, tarif_id, models_num
        waf_id, tarif_id, models_num, geo_token = get_auth(key, agent_ip)

        # WAF ID check
        if not waf_id:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Processing JSON
        ##

        # init
        resp = {'status': 'success', 'deleted': []}

        try:

            # insert the data
            query = """
                DELETE FROM
                    events
                WHERE (
                    waf_id = %s
                        AND
                    timestamp <= %s
                        AND
                    CAST (id AS TEXT) SIMILAR TO %s
                        AND
                    ip SIMILAR TO %s
                        AND
                    uuid SIMILAR TO %s
                        AND
                    agent SIMILAR TO %s
                        AND
                    level SIMILAR TO %s
                        AND
                    category SIMILAR TO %s
                )
                RETURNING id
            """
            cursor.execute(
                query, (
                    waf_id,
                    jdata['timestamp'],
                    jdata['id'],
                    jdata['ip'],
                    jdata['uuid'],
                    jdata['agent'],
                    jdata['level'],
                    jdata['category'],
                )
            )
            res = cursor.fetchall()

            # processing result
            resp['deleted'] = len(res) if res else 0

        except Exception as e:
            log.error('An error occurred while deleting the events: {}'.format(e))
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': str(e)})
            return Response(resp, status=200)

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(json.dumps(resp), status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


##
# Scanner settings
##

@app.route(get_nws_target_uri, methods=['POST'])
def get_nws_target_f():

    # set URL
    url = request.path

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=200)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id
        waf_id, tarif_id, _, _ = get_auth(key, agent_ip)

        # WAF ID and tarif_id check
        if not waf_id or tarif_id < 2:
            return Response(status=401)

        ##
        # Extract data from the database
        ##

        # init
        targets_list = []
        response = {'status': 'success', 'targets': targets_list}
        order_by = 'ORDER BY target'

        # search data
        query = """ SELECT to_jsonb(array_agg(nws_target {})) FROM nws_target WHERE (waf_id = %s) """.format(order_by)
        cursor.execute(query, (waf_id,))
        res = cursor.fetchone()[0]

        if res:

            # update targets
            for r in res:

                # del columns
                del r['waf_id']

                # dict/list normalization
                for k in ['auth_params', 'exclude_modules']:
                    if k in r and r[k]:
                        r[k] = ast.literal_eval(r[k])

                # update targets list
                targets_list.append(r)

            # update the response
            response['targets'] = targets_list

        # close database connection
        close_db_connection(connect, cursor)

        # response
        return Response(json.dumps(response), status=200)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(set_nws_target_uri, methods=['POST'])
def set_nws_target_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id
        waf_id, tarif_id, _, _ = get_auth(key, agent_ip)

        # WAF ID and tarif_id check
        if not waf_id or tarif_id < 2:
            return Response(status=401)

        ##
        # Processing request data
        ##

        # init
        k_list = []
        v_list = []

        # extract the target
        k = 'target'
        if k in jdata and isinstance(jdata[k], str) and jdata[k]:
            target = jdata[k]
        else:
            log.error('Required key "{}" is missing or incorrect'.format(k))
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        # processing data
        for k in nws_target_cols:

            # auth_param processing
            if k == 'auth_params' and k in jdata:

                # convert all keys to lowercase and exclude unknown keys
                try:
                    jdata[k] = {ik.lower(): iv for ik, iv in jdata[k].items() if ik in nws_target_ap_cols}
                except Exception as e:
                    log.error(
                        'An error occurred while processing {}: {} ({})'
                        .format(k, e, jdata)
                    )

            # processing data
            if k in jdata:

                # value validations
                status = nws_target_value_validation(agent_ip, k, jdata[k])

                # check status
                if not status:

                    # response
                    close_db_connection(connect, cursor)
                    err_msg = str('Incorrect value for {}'.format(k))
                    resp = json.dumps({'status': 'fail', 'description': err_msg})
                    return Response(resp, status=201)

                # replace quotes in list
                if isinstance(jdata[k], dict) or isinstance(jdata[k], list):
                    v = str(jdata[k])
                else:
                    v = jdata[k]

                # convert to srt and replace newline/multiple space to space
                if isinstance(v, str):
                    v = re.sub(r'\s+', ' ', v)

                # update key/value lists
                k_list.append(k)
                v_list.append(v)

        ##
        # Save data to the database
        ##

        # check JSON parameters
        if not k_list:
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'Parameters are missing in JSON'})
            return Response(resp, status=201)

        try:

            try:

                # insert row if it does not exist
                query = """ INSERT INTO nws_target (target, waf_id) VALUES (%s, %s) ON CONFLICT DO NOTHING """
                cursor.execute(query, (target, waf_id,))

                # update data
                query = """ UPDATE nws_target SET ({}) = %s WHERE (target = %s AND waf_id = %s) """.format(','.join(k_list))
                cursor.execute(query, (tuple(v_list), target, waf_id,))

            except Exception as e:

                # logging
                log_key = 'x' * 26 + key[-6:]
                log.info('An error occurred while processing data for key "{}" in nws_target: {}'.format(log_key, e))

                # response
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

            # close database connection
            close_db_connection(connect, cursor)

            # response
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:

            # response
            resp = json.dumps({'status': 'fail', 'description': str(e)})
            return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)


@app.route(del_nws_target_uri, methods=['POST'])
def del_nws_target_f():

    # set URL
    url = request.path

    try:

        # RW DB type only
        if ro_mode:
            return Response(status=502)

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # extract client IP
        agent_ip = get_agent_ip(request)

        # extract data from the request
        try:
            jdata = json.loads(request.data)
        except Exception:
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'The data is not in JSON format'})
            return Response(resp, status=201)

        # convert all keys to lowercase
        try:
            jdata = {k.lower(): v for k, v in jdata.items()}
        except Exception as e:
            log.error(
                'An error occurred while converting keys to lowercase from {} to {}: {} ({})'
                .format(agent_ip, url, e, jdata)
            )

        # extract lic. key
        try:
            key = jdata['key']
        except Exception as e:
            log.error(
                'An error occurred while extracting product key from {} to {}: {}'
                .format(agent_ip, url, e)
            )
            close_db_connection(connect, cursor)
            return Response(status=401)

        # get waf_id, tarif_id
        waf_id, tarif_id, _, _ = get_auth(key, agent_ip)

        # WAF ID and tarif_id check
        if not waf_id or tarif_id < 2:
            close_db_connection(connect, cursor)
            return Response(status=401)

        ##
        # Processing request data
        ##

        # extract the name
        k = 'target'
        if k in jdata and isinstance(jdata[k], str) and jdata[k]:
            target = jdata[k]
        else:
            log.error('Required key "{}" is missing or incorrect'.format(k))
            close_db_connection(connect, cursor)
            resp = json.dumps({'status': 'fail', 'description': 'Required key missing'})
            return Response(resp, status=201)

        try:

            try:

                # delete data
                query = """ DELETE FROM nws_target WHERE (target = %s AND waf_id = %s) """
                cursor.execute(query, (target, waf_id,))

            except Exception as e:

                # logging
                log_key = 'x' * 26 + key[-6:]
                log.info('An error occurred while processing data for key "{}" in nws_target: {}'.format(log_key, e))

                # response
                close_db_connection(connect, cursor)
                resp = json.dumps({'status': 'fail', 'description': str(e)})
                return Response(resp, status=201)

            # close database connection
            close_db_connection(connect, cursor)

            # response
            resp = json.dumps({'status': 'success'})
            return Response(resp, status=201)

        except Exception as e:

            # response
            resp = json.dumps({'status': 'fail', 'description': str(e)})
            return Response(resp, status=201)

    except Exception as e:
        log.error('An error occurred in {}: {}'.format(url, e))

    # default response
    return Response(status=503)
