#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2024 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import base64
import json
import logging
import psycopg2
import re
import requests

from decouple import Config, RepositoryEnv
from logging.handlers import RotatingFileHandler
from multiprocessing import Lock
from netaddr import IPAddress
from pymemcache.client.base import Client
from time import sleep

##
# URIs
##

# Remote server URI
auth_uri = 'https://nemesida-security.com/nw/auth/v3'
rl_desc_uri = 'https://nemesida-security.com/nw/rl-desc/'
geoip_batch_uri = 'https://geoip.nemesida-security.com/geoip/v3/batch?geo_token='
geocc_memc_key = 'geocc_'

# Local server URI
base_api_uri = '/nw-api'

# Attacks, scans URLs
set_attack_uri = base_api_uri + '/attack'
get_attack_uri = base_api_uri + '/get_attack'
get_attack_stats_uri = base_api_uri + '/get_attack_stats'
set_scanner_uri = base_api_uri + '/scan'

# Misc.
headers = {'User-agent': 'Nemesida WAF API'}
waf_id_ce = 'COMMUNITYEDITION'

##
# Log settings
##

logf = '/var/log/uwsgi/nw-api/nw-api-logging.log'
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)
fh = logging.handlers.RotatingFileHandler(logf, maxBytes=50 * 1024 * 1024, backupCount=5)
formatter = logging.Formatter('[%(asctime)s] %(levelname)s - %(message)s', '%m-%d %H:%M:%S')
fh.setFormatter(formatter)
log.addHandler(fh)

##
# Export settings
##

# init
proxy = {}
settings_file = '/var/www/nw-api/settings.py'

try:

    # init config
    env_config = Config(RepositoryEnv(settings_file))

    # init database settings
    db_host = env_config.get('DB_HOST', default='127.0.0.1')
    db_port = env_config.get('DB_PORT', default='5432')
    db_name = env_config.get('DB_NAME', default='waf')
    db_user = env_config.get('DB_USER', default='nw_api')
    db_pass = env_config.get('DB_PASS', default='')
    ro_mode = env_config.get('RO_MODE', default=False, cast=bool)

    # database settings processing
    if not db_pass:
        log.warning(
            'Option "db_pass" is not set, make changes to the configuration file and restart the services'
        )
        exit()

    # init memcached
    memc_lock = Lock()
    memcached_host = env_config.get('MEMCACHED_HOST', default='127.0.0.1')
    memcached_port = env_config.get('MEMCACHED_PORT', default='11211')
    memc = Client((memcached_host, int(memcached_port)))

    # init proxy
    proxy_settings = env_config.get('PROXY', default='')

    # proxy settings processing
    if proxy_settings:
        try:

            # init
            proxy = proxy_settings

            # set schema if not exists
            if not re.match(r'^https?://', proxy_settings.lower()):
                proxy = 'http://' + proxy_settings

            # logging
            log.info('Proxy server used: {}'.format(proxy))

            # set proxy
            proxy = {'http': proxy, 'https': proxy}

        except Exception as err:
            log.error(
                'An error occurred while adding the schema for proxy-server: {}'.format(err)
            )

except Exception as err:
    log.error(
        'An error occurred while exporting settings from file {}: {}'
        .format(settings_file, err)
    )
    exit()

##
# Functions
##


def init_db_connection():
    try:

        connect = psycopg2.connect(
            database=db_name,
            user=db_user,
            password=db_pass,
            host=db_host,
            port=db_port,
            connect_timeout=10
        )
        connect.set_session(autocommit=True)
        return connect

    except Exception as e:
        log.error('An error occurred while connecting to the database: {}'.format(e))
        return False


def close_db_connection(connect, cursor):
    try:
        cursor.close()
        connect.close()
    except Exception as e:
        log.error('An error occurred while closing database connection: {}'.format(e))


def get_auth(key, agent_ip):

    # init
    waf_id = ''
    tarif_id = 0
    models_num = 0
    geo_token = ''

    # init memcached
    memc_key = 'auth_{}'.format(key)

    ##
    # Search data in memcached
    ##

    try:

        # get data from memcached
        with memc_lock:
            memc_bin = memc.get(memc_key)

        # if data in memcached
        if memc_bin:

            # processing memcached data
            data = memc_bin.decode('UTF-8')
            data = json.loads(data)

            # If lic. key is valid
            if data['status'] == 'success':

                # extract data
                waf_id = str(data['waf_id'])
                tarif_id = data['tarif_id']
                models_num = data['models_num']
                geo_token = data['geo_token']

        # return
        if waf_id:
            return waf_id, tarif_id, models_num, geo_token

    except Exception as e:
        log.error('An error occurred while processing WAF ID from the memcached: {}'.format(e))

    ##
    # If WAF ID not in memcached
    # receiving from the Nemesida WAF Auth
    ##

    while True:
        try:

            # init
            s = init_session()

            # send request
            result = s.post(
                auth_uri,
                headers=headers,
                proxies=proxy,
                data={'key': key},
                verify=True,
                timeout=15
            )
            status_code = result.status_code

            # processing
            if status_code == 201:
                try:

                    # init
                    data = json.loads(result.text)

                    # extract data
                    waf_id = str(data['waf_id'])
                    tarif_id = data['tarif_id']
                    models_num = data['models_num']
                    geo_token = data['geo_token']

                    # update WAF ID in memcached
                    try:
                        if waf_id:
                            with memc_lock:
                                memc.set(memc_key, json.dumps(data), 600)
                    except Exception as e:
                        log_key = 'x' * 26 + key[-6:]
                        log.error(
                            'An error occurred while storing auth data for product key {} to the memcached from client {}: {}'
                            .format(log_key, agent_ip, e)
                        )

                    # return
                    return waf_id, tarif_id, models_num, geo_token

                except Exception as e:
                    log_key = 'x' * 26 + key[-6:]
                    log.error(
                        'An error occurred while processing auth data for product key {} from client {}: {}'
                        .format(log_key, agent_ip, e)
                    )
                    sleep(1)

            elif status_code == 401:
                log_key = 'x' * 26 + key[-6:]
                log.warning('Invalid product key {} received from client {}'.format(log_key, agent_ip))
                return waf_id, tarif_id, models_num, geo_token

        except Exception as e:
            log.error('An error occurred while processing the request from client {}: {}'.format(agent_ip, e))


def get_waf_id(key, agent_ip):
    try:
        waf_id, _, _, _ = get_auth(key, agent_ip)
        return waf_id
    except Exception as e:
        log.error('An error occurred while processing WAF ID: {}'.format(e))
        return False


def init_db():

    ##
    # Create new column if not exist
    ##

    try:

        log.info('The database is being initialized')

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        """
        Rename the sequence
        """

        # attack_ids
        query = """ ALTER SEQUENCE IF EXISTS attack_id_seq RENAME TO attack_ids """
        cursor.execute(query,)

        # scan_report_ids
        try:
            query = """ ALTER SEQUENCE IF EXISTS scan_report_id_seq RENAME TO scan_report_ids """
            cursor.execute(query,)
        except Exception as e:
            log.error('An error occurred while renaming sequence scan_report_id_seq: {}'.format(e))
            try:
                query = """ DROP SEQUENCE IF EXISTS scan_report_id_seq """
                cursor.execute(query,)
            except Exception as e:
                log.error('An error occurred while deleting sequence scan_report_id_seq: {}'.format(e))
                try:
                    query = """ DROP SEQUENCE IF EXISTS scan_report_ids """
                    cursor.execute(query,)
                    query = """ ALTER SEQUENCE IF EXISTS scan_report_id_seq RENAME TO scan_report_ids """
                    cursor.execute(query,)
                except Exception as e:
                    log.error('An error occurred while deleting sequence scan_report_ids: {}'.format(e))

        """
        Create new tables if not exists
        """

        fname = 'api.sql'
        log.info('Database processing using {}'.format(fname))
        cursor.execute(open(fname, 'r').read())
        log.info('The database has been processed successfully using {}'.format(fname))

        ##
        # Updating the tables
        ##

        log.info('Updating the database tables')

        """
        Table: attack
        """

        # Receive columns name
        query = """
            SELECT
                column_name
            FROM
                INFORMATION_SCHEMA.COLUMNS
            WHERE
                TABLE_NAME = 'attack'
            """
        cursor.execute(query,)
        res = cursor.fetchall()

        # rename content to body
        if res and ('content',) in res:
            query = """ ALTER TABLE attack RENAME COLUMN content TO body """
            cursor.execute(query,)

        # rename uri to url
        if res and ('uri',) in res:
            query = """ ALTER TABLE attack RENAME COLUMN uri TO url """
            cursor.execute(query,)

        # modify the columns
        query = """
            ALTER TABLE
                attack
            DROP COLUMN IF EXISTS
                bot,
            ADD COLUMN IF NOT EXISTS
                group_id text,
            ADD COLUMN IF NOT EXISTS
                blocked boolean,
            ADD COLUMN IF NOT EXISTS
                possible boolean,
            ADD COLUMN IF NOT EXISTS
                cc char(2) DEFAULT NULL,
            ALTER COLUMN
                url DROP NOT NULL,
            ALTER COLUMN
                param DROP NOT NULL,
            ALTER COLUMN
                other_headers DROP NOT NULL,
            ALTER COLUMN
                referer DROP NOT NULL,
            ALTER COLUMN
                mz DROP NOT NULL,
            ALTER COLUMN
                description DROP NOT NULL,
            ALTER COLUMN
                rule_id DROP NOT NULL,
            ALTER COLUMN
                rule_id SET DEFAULT 0,
            ADD COLUMN IF NOT EXISTS
                long_description text
        """
        cursor.execute(query,)

        # drop indexes
        query = """ DROP INDEX IF EXISTS idx_mtp """
        cursor.execute(query,)

        log.info('Table "attack" has been successfully updated or no migration is required')

        """
        Table: scan_report
        """

        ##
        # Add new column mz
        ##

        query = """
            ALTER TABLE
                scan_report
            ADD COLUMN IF NOT EXISTS
                mz text
        """
        cursor.execute(query,)

        # drop NOT NULL for specific columns
        query = """
            ALTER TABLE
                scan_report
            ALTER COLUMN
                param DROP NOT NULL,
            ALTER COLUMN
                payload DROP NOT NULL,
            ALTER COLUMN
                type DROP NOT NULL,
            ALTER COLUMN
                url DROP NOT NULL,
            ALTER COLUMN
                data DROP NOT NULL
        """
        cursor.execute(query,)

        log.info('Table "scan_report" has been successfully updated or no migration is required')

        """
        Table: dyn_settings
        """

        # add new column nwaf_b64_decode_disable
        query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_b64_decode_disable text """
        cursor.execute(query,)

        # add new column nwaf_geoip_mla_disable
        query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_geoip_mla_disable text """
        cursor.execute(query,)

        # add new column nwaf_rfc_violation_wl
        query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_rfc_violation_wl text """
        cursor.execute(query,)

        # add new column nwaf_rfc_violation_lm
        query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_rfc_violation_lm text """
        cursor.execute(query,)

        # add new column nwaf_body_bin_exclude
        query = """ ALTER TABLE dyn_settings ADD COLUMN IF NOT EXISTS nwaf_body_bin_exclude text """
        cursor.execute(query,)

        log.info('Table "dyn_settings" has been successfully updated or no migration is required')

        """
        Table: openapi_url
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_url
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)

        log.info('Table "openapi_url" has been successfully updated or no migration is required')

        """
        Table: openapi_schemas
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_schemas
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)

        # modify the openapi_url_idx index
        try:

            query = """
                SELECT
                    indexdef
                FROM
                    pg_indexes
                WHERE
                    schemaname = 'public'
                        AND
                    indexname = 'openapi_url_idx'
                        AND
                    (
                        indexdef LIKE '%USING btree (waf_id)'
                            OR
                        indexdef LIKE '%USING btree (waf_id, url)'
                    )
            """

            cursor.execute(query,)
            if cursor.fetchone():

                query = """ DROP INDEX IF EXISTS new_openapi_url_idx """
                cursor.execute(query,)

                query = """ CREATE UNIQUE INDEX new_openapi_url_idx ON openapi_url USING btree (waf_id, md5(url)) """
                cursor.execute(query,)

                query = """ DROP INDEX openapi_url_idx """
                cursor.execute(query,)

                query = """ ALTER INDEX new_openapi_url_idx RENAME TO openapi_url_idx """
                cursor.execute(query,)

        except Exception as e:
            log.error('An error occurred while processing openapi_url_idx: {}'.format(e))

        log.info('Table "openapi_schemas" has been successfully updated or no migration is required')

        """
        Table: openapi_schema_paths
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_schema_paths
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)

        # modify the openapi_schema_paths_idx index
        try:

            query = """
                SELECT
                    indexdef
                FROM
                    pg_indexes
                WHERE
                    schemaname = 'public'
                        AND
                    indexname = 'openapi_schema_paths_idx'
                        AND
                    indexdef LIKE '%USING btree (waf_id, url, schema_name)'
            """

            cursor.execute(query,)
            if cursor.fetchone():

                query = """ DROP INDEX IF EXISTS new_openapi_schema_paths_idx """
                cursor.execute(query,)

                query = """ CREATE UNIQUE INDEX new_openapi_schema_paths_idx ON openapi_schema_paths USING btree (waf_id, md5(url), schema_name) """
                cursor.execute(query,)

                query = """ DROP INDEX openapi_schema_paths_idx """
                cursor.execute(query,)

                query = """ ALTER INDEX new_openapi_schema_paths_idx RENAME TO openapi_schema_paths_idx """
                cursor.execute(query,)

        except Exception as e:
            log.error('An error occurred while processing the openapi_schema_paths_idx: {}'.format(e))

        # logging
        log.info('Table "openapi_schemas_path" has been successfully updated or no migration is required')

        """
        Table: openapi_exclude_paths
        """

        # modify the columns
        query = """
            ALTER TABLE
                openapi_exclude_paths
            ALTER COLUMN
                timestamp TYPE timestamp without time zone USING timestamp::timestamp without time zone
        """
        cursor.execute(query,)

        log.info('Table "openapi_exclude_paths" has been successfully updated or no migration is required')

        """
        Table: ml
        """

        # drop table
        query = """ DROP TABLE IF EXISTS ml """
        cursor.execute(query,)

        log.info('Table "ml" has been successfully dropped or no migration is required')

        """
        Table: nginx_global
        """

        # modify the columns
        query = """
            ALTER TABLE
                nginx_global
            ALTER COLUMN
                resolver_timeout TYPE text,
            ALTER COLUMN
                client_max_body_size TYPE text
        """
        cursor.execute(query,)

        log.info('Table "nginx_global" has been successfully dropped or no migration is required')

        """
        Table: nginx_vhost
        """

        # modify the columns
        query = """
            ALTER TABLE
                nginx_vhost
            ALTER COLUMN
                ssl_session_timeout TYPE text,
            ADD COLUMN IF NOT EXISTS
                add_header text,
            ADD COLUMN IF NOT EXISTS
                error_page text
        """
        cursor.execute(query,)

        log.info('Table "nginx_vhost" has been successfully dropped or no migration is required')

        ##

        # modify the columns
        query = """
            ALTER TABLE
                mt_main
            ALTER COLUMN
                cc TYPE char(2),
            ALTER COLUMN
                cc SET DEFAULT NULL
        """
        cursor.execute(query,)

        log.info('Table "mt_main" has been successfully dropped or no migration is required')

        ##

        log.info('The database initialization is complete')

        # close database connection
        connect.close()

    except Exception as e:
        log.error('An error occurred while updating the table structure: {}'.format(e))


def init_session():

    s = requests.Session()
    s.trust_env = False
    return s


def get_rldesc(rule_id):

    # init
    rldesc = ''
    rule_id = str(rule_id)

    try:

        # logging
        log.info('Updating description for RuleID {}'.format(rule_id))

        # init
        s = init_session()
        data = {'rule_id': rule_id}
        data = json.dumps(data)

        # receiving rule description
        state = s.post(
            rl_desc_uri,
            proxies=proxy,
            headers=headers,
            data=data,
            timeout=120
        )
        status_code = state.status_code

        # processing result
        if status_code != 200 and status_code != 201:
            log.error(
                'Incorrect server answer for RuleID {}: {}, text: {}'.format(rule_id, state.status_code, state.text)
            )
            return rldesc

        desc_dict = state.text
        desc_dict = json.loads(desc_dict)

        return desc_dict['desc']

    except Exception as e:
        log.error(
            'An error occurred in while processing description for rule {}: {}'.format(rule_id, e)
        )
        return rldesc


def get_geocc_remote(ips, waf_id):
    try:

        # skip empty/CE WAF ID
        if not waf_id.strip() or waf_id == waf_id_ce:
            return

        # get GeoToken
        geo_token = get_geo_token(waf_id)
        if not geo_token:
            return

        ##
        # Receive GeoIP from the remote server
        # if data not found in the memcached and IP is a public
        ##

        # init
        s = init_session()

        # receiving GeoIP
        result = s.post(
            geoip_batch_uri + geo_token,
            proxies=proxy,
            headers=headers,
            data=str(ips),
            verify=True,
            timeout=30,
        )

        # processing status code
        if not re.match(r'^20', str(result.status_code)):
            log.error(
                'An error occurred while processing data from GeoIP server: {}'
                .format(result)
            )
            return

        ##
        # JSON processing
        ##

        # load to JSON
        try:
            data = json.loads(result.text)
        except Exception as e:
            log.error(
                'An error occurred while processing data from GeoIP server: {} ({})'
                .format(e, result.text)
            )
            return

        # if result is not empty/False
        if data:

            # update data in the memcached
            for k, v in data.items():

                # set the memcached key
                memc_key = geocc_memc_key + str(k)

                # skip private
                if v == 'Private IP':
                    with memc_lock:
                        memc.set(memc_key, v, 86400)  # 86400 - 1 d.

                else:

                    # load value from JSON
                    try:
                        d = json.loads(v)
                    except Exception as e:
                        logging.error('An error occurred while processing data for {}: {}'.format(v, e))
                        continue

                    # skip non-success
                    if d.get('status', None) != 'success':
                        continue

                    # update data in the memcached
                    try:
                        if 'cc' in d and d['cc']:
                            with memc_lock:
                                memc.set(memc_key, d['cc'], 86400)  # 86400 - 1 d.
                    except Exception as e:
                        logging.error('An error occurred while storing data for {}: {}'.format(k, e))

    except Exception as e:
        log.error('An error occurred in get_geocc_remote: {}'.format(e))


def get_geocc_local(ip):
    try:

        # init memcached
        memc_key = geocc_memc_key + str(ip)

        # get CC from the memcached
        try:

            # search data in memcached
            with memc_lock:
                memout = memc.get(memc_key)

            # if data in memcached
            if memout:
                return memout.decode('UTF-8')

        except Exception as e:
            log.error('An error occurred while processing GeoIP data from memcached: {}'.format(e))

    except Exception as e:
        log.error('An error occurred in get_geocc_local: {}'.format(e))

    # default action
    return None


def geo_token_update(key, agent_ip):
    try:
        waf_id, _, _, geo_token = get_auth(key, agent_ip)
        put_geo_token_to_db(waf_id, geo_token)
    except Exception as e:
        log.error('An error occurred while processing GeoToken: {}'.format(e))


def get_geo_token(waf_id):
    try:

        # search GeoToken in the database
        query = """ SELECT geo_token FROM geo_tokens WHERE waf_id = %s """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id,))
        row = cursor.fetchone()
        connect.close()

        # processing result
        return row[0] if row else ''

    except Exception as e:
        log.error(
            'An error occurred while processing GeoToken for WAF ID {} in the database: {}'.format(waf_id, e)
        )
        return ''


def put_geo_token_to_db(waf_id, geo_token):
    try:

        # store GeoToken to the database
        query = """ INSERT INTO geo_tokens (waf_id, geo_token) VALUES (%s, %s) ON CONFLICT DO NOTHING """
        connect = init_db_connection()
        cursor = connect.cursor()
        cursor.execute(query, (waf_id, geo_token,))
        connect.close()

    except Exception as e:
        log.error('An error occurred during storing data to {}: {}'.format(geo_token, e))


def get_agent_ip(request):

    # extract client IP
    try:
        return request.remote_addr
    except Exception:
        return ''


def set_attack_processing(data, agent_ip):
    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        ##
        # Processing dictionary of the attacks
        ##

        for item in data:
            try:

                # extract lic. key
                key = item.get('waf_lic_key', '')

                # WAF ID processing
                if key == 'none':
                    waf_id = waf_id_ce
                else:
                    waf_id = get_waf_id(key, agent_ip)
                    geo_token_update(key, agent_ip)
                    if not waf_id:
                        continue

                # init
                fields_dict = {}

                # description fix
                if 'description' not in item:
                    item['description'] = ''

                # long_description fix
                if 'long_description' not in item:
                    item['long_description'] = ''

                # group_id fix
                if 'group_id' not in item:
                    item['group_id'] = ''

                # url fix
                if 'uri' in item:
                    item['url'] = item['uri']
                    item.pop('uri')

                # rename content to boy
                try:
                    if 'content' in item:
                        item['body'] = item.pop('content')
                except Exception as e:
                    log.error(
                        'An error occurred while renaming column "content" to "body": {}'.format(e)
                    )

                ##
                # b64decode
                ##

                # init
                b64_fields = (
                    'ip',
                    'request_id',
                    'vhost',
                    'url',
                    'param',
                    'mz',
                    'method',
                    'ua',
                    'referer',
                    'other_headers',
                    'body',
                    'cookie',
                    'description',
                    'long_description',
                    'group_id',
                )

                # decode custom field
                for field in b64_fields:
                    try:
                        # base64 decode
                        fields_dict[field] = base64.b64decode(item[field]).decode('UTF-8', 'ignore')
                    except Exception as e:
                        log.error(
                            'An error occurred while decoding the field (client IP: {}): {}'
                            .format(agent_ip, e)
                        )
                        fields_dict[field] = 'API DECODE ERROR'

                ##
                # Extract data from the JSON item
                ##

                # extract items
                bt = int(item.get('bt', 0))
                rule_id = int(item.get('rule_id', 0))
                ip = fields_dict.get('ip', '')
                vhost = fields_dict.get('vhost', '')
                mz_list = fields_dict.get('mz', '').split('|')
                request_id = fields_dict.get('request_id', '0')

                # mz normalization
                mz_list = attack_zone_replace(mz_list)

                # extract body
                try:
                    body = bytearray(fields_dict.get('body', ''), 'UTF-8')
                except Exception as e:
                    log.error(
                        'An error occurred while extracting the "body" data from request {}: {}'.format(request_id, e)
                    )
                    body = ''

                ##
                # Processing for each zone in mz
                ##

                for mz in mz_list:
                    try:

                        # init
                        description = ''

                        # processing specific bt
                        if bt in [1, 2] and rule_id != 0:

                            ##
                            # Receive description
                            ##

                            # init
                            memc_key = 'rldsc_' + str(rule_id)

                            # Search description in the memcached
                            try:

                                # get data from memcached
                                with memc_lock:
                                    memc_bin = memc.get(memc_key)

                                # if data exist in memcached
                                if memc_bin:
                                    description = memc_bin.decode('UTF-8')

                            except Exception as e:
                                log.error(
                                    'An error occurred while processing rule description from the memcached: {}'
                                    .format(e)
                                )
                                description = ''

                            # Receive the description from the remote server if description is empty
                            if not description or description == '-':

                                # receive the description
                                description = get_rldesc(rule_id)

                                # update data in memcached
                                try:
                                    if description:
                                        with memc_lock:
                                            memc.set(memc_key, description, 86400)  # 84600 - 1 day
                                except Exception as e:
                                    log.error(
                                        'An error occurred while storing description "{}" for rule ID {}'
                                        ' in the memcached: {}'
                                        .format(description, rule_id, e)
                                    )

                        elif bt == 4:
                            description = 'Virus (' + fields_dict['description'] + ')'

                        elif bt == 5:
                            mz = ''
                            description = fields_dict['description']

                        elif bt == 7:
                            description = fields_dict['description']
                            if not description or description == '-':
                                description = 'Brute-force'

                        elif bt == 9:
                            description = fields_dict['description']
                            if not description or description == '-':
                                description = 'Flood'

                        elif bt == 10:
                            mz = ''
                            description = fields_dict['description']

                        else:
                            description = fields_dict['description']
                            if not description:
                                description = ''

                        ##
                        # Store attack to the database
                        ##

                        try:

                            # preparing a query
                            query = """
                                INSERT INTO attack (
                                    timestamp, lm, bt, mz, ip, vhost, rule_id, request_id,
                                    description, long_description, waf_id, url, method,
                                    referer, param, ua, cookie, other_headers, body,
                                    group_id, blocked, possible
                                ) VALUES (
                                    to_timestamp(%s), %s, %s, %s, %s, %s, %s,
                                    %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                                    %s, %s, %s, %s
                                ) ON CONFLICT DO NOTHING
                           """

                            ar = (
                                item.get('timestamp', ''),
                                item.get('lm', ''),
                                bt,
                                mz,
                                ip,
                                vhost,
                                rule_id,
                                request_id,
                                description,
                                fields_dict['long_description'],
                                waf_id,
                                fields_dict['url'],
                                fields_dict['method'],
                                fields_dict['referer'],
                                fields_dict['param'],
                                fields_dict['ua'],
                                fields_dict['cookie'],
                                fields_dict['other_headers'],
                                body,
                                fields_dict['group_id'],
                                item.get('blocked', None),
                                item.get('possible', None),
                            )

                            # log.info('Preparing the query for writing to the database: {}'.format(ar))

                            try:

                                # insert data to the DB
                                cursor.execute(query, ar)

                                # logging
                                log.info(
                                    'Request {} (BT: {}, client: {}, server: {}, WAF ID: {}) received from {}'
                                    .format(request_id, bt, ip, vhost, waf_id, agent_ip)
                                )

                            except Exception as e:
                                log.error(
                                    'An error occurred while storing data in database from request'
                                    ' {} (BT: {}, client: {}, server: {}, WAF ID: {}): {} ({})'
                                    .format(request_id, bt, ip, vhost, waf_id, e, ar)
                                )

                        except Exception as e:
                            log.error(
                                'An error occurred during inserting request {}'
                                ' (BT: {}, client: {}, server: {}, WAF ID: {}) to the database: {}'
                                .format(request_id, bt, ip, vhost, waf_id, e)
                            )

                    except Exception as e:
                        log.error('An error occurred while processing zone {} in {}: {}'.format(mz, request_id, e))

            except Exception as e:
                log.error('An error occurred while processing JSON item: {} ({})'.format(e, item))

        # close database connection
        close_db_connection(connect, cursor)

    except Exception as e:
        log.error('An error occurred while processing data in set_attack_processing: {}'.format(e))


def attack_zone_replace(mz):
    try:

        # init
        result = []

        # replace conditions
        repl = {
            "uri": "URL",
            "args": "ARGS",
            "body": "BODY",
            "ua": "User-Agent",
            "cookie": "Cookie",
            "referer": "Referer",
            "other_headers": "HEADERS",
        }

        # processing for each zone in mz list
        for z in mz:

            # zone normalization
            for k, v in repl.items():
                z = z.replace(k, v)

            # update result list
            result.append(z)

        return result

    except Exception as e:
        log.error('An error occurred while mz normalization: {} ({})'.format(e, mz))
        return mz


def vhost_normalization(vhost):
    try:

        # init
        dot = r'^\.'
        sub = r'^\*\.'

        # *
        if vhost == '*':
            vhost = 'all'

        # .example.com
        elif re.search(dot, vhost):
            vhost = re.sub(dot, '', vhost)

        # *.example.com
        elif re.search(sub, vhost):
            vhost = re.sub(sub, 'sub_', vhost)

    except Exception as e:
        log.error('An error occurred while vhost normalization (vhost: {}): {}'.format(vhost, e))

    return vhost


def idna_vhost_list_normalization(vhost):
    try:

        # dot
        if vhost.startswith('.'):

            # processing for non-latin vhost
            try:
                vhost = '.' + vhost[1:].encode('idna').decode('UTF-8')
            except Exception as e:
                log.error('An error occurred while IDNA normalization for virtual host {}: {}'.format(vhost, e))

        # other
        else:

            # processing for non-latin vhost
            try:
                vhost = vhost.encode('idna').decode('UTF-8')
            except Exception as e:
                log.error('An error occurred while IDNA normalization for virtual host {}: {}'.format(vhost, e))

    except Exception as e:
        log.error('An error occurred while normalizing vhost {}: {}'.format(vhost, e))

    return vhost


def get_data_from_json(request, url):

    # extract client IP
    agent_ip = get_agent_ip(request)

    # load JSON
    try:
        return json.loads(request.data)
    except Exception as e:
        log.error(
            'An error occurred while processing the request to {} (client IP: {}): {}'
            .format(url, agent_ip, e)
        )
        return None


def is_base64(s):

    ##
    # Check is data in base64
    ##

    # data is not in Base64 if contains only numbers
    if re.match(r'^\d+={,2}$', s):
        return False

    # check is data in Base64
    try:
        return base64.b64encode(base64.b64decode(s.encode())).decode('UTF-8', 'ignore') in [s, s + '=', s + '==']
    except Exception:
        return False


def is_private(ip):
    try:

        # init
        addr = IPAddress(ip)

        # netaddr >= 1.1.0
        if 'is_ipv4_private_use' in dir(addr):

            # multicast IP
            if addr.is_multicast():
                return True

            # if not global IP
            if not addr.is_global():
                return True

        else:

            # non-public IP
            if addr.is_private() or addr.is_loopback() or addr.is_multicast():
                return True

        # loopback/multicast validation
        if addr.is_loopback() or addr.is_multicast():
            return True

    except Exception as e:
        log.error('An error occurred while processing IP {}: {}'.format(ip, e))

    return False


def cc_upd_db(table):

    ##
    # Processing CC in attack
    ##

    try:

        # init database connection
        connect = init_db_connection()
        cursor = connect.cursor()

        # get the data
        query = """
            SELECT DISTINCT
                ip, waf_id
            FROM
                {}
            WHERE (
                cc IS NULL
                    AND
                waf_id != %s
                    AND
                waf_id != ''
            )
        """.format(table)
        cursor.execute(query, (waf_id_ce,))
        res = cursor.fetchall()

        # skip empty
        if not res:
            return

        # preprocessing the result
        waf_ids = set(x[1] for x in res if x)

        ##
        # Updating CC
        ##

        # processing IPs for each WAF ID
        for waf_id in waf_ids:

            # init
            private_ips = []

            # extrack IPs
            all_ips = set(x[0] for x in res if x and x[1] == waf_id)

            # get private IPs
            for ip in all_ips:
                if is_private(ip):
                    private_ips.append(ip)

            # remove private IPs from the public IPs
            public_ips = [x for x in all_ips if x not in private_ips]

            # split the list into chunks
            chunks = [public_ips[x:x + 100] for x in range(0, len(public_ips), 100)]

            # processing the chunk list
            for chunk in chunks:

                # put GeoIP CC to the memcached
                get_geocc_remote(chunk, waf_id)

            # update CC for each IP in the database
            for ip in all_ips:

                # private IP
                if ip in private_ips:
                    cc = ''

                else:

                    # get the CC
                    cc = get_geocc_local(ip)

                    # skip empty
                    if cc is None:
                        continue

                    # private IP in the memcached
                    if cc == 'Private IP':
                        cc = ''

                # update
                query = """ UPDATE {} SET cc = %s WHERE (ip = %s AND waf_id = %s) """.format(table)
                cursor.execute(query, (cc, ip, waf_id,))

        # close database connection
        close_db_connection(connect, cursor)

    except Exception as e:
        log.error('An error occurred while updating CountryCode in "{}" table: {}'.format(table, e))
