#!/usr/bin/env python3

"""
Nemesida WAF API
Copyright (c) 2017-2024 Pentestit LLC - All Rights Reserved
Author: Romanov R.
"""

import hashlib
import json
import jsonref

from psycopg2.extras import RealDictCursor

from core import base_api_uri
from core import close_db_connection
from core import init_db_connection
from core import log
from core import memc
from core import memc_lock

##
# URLs
##

set_openapi_schema_uri = base_api_uri + '/set_openapi_schema'
get_openapi_schemas_uri = base_api_uri + '/get_openapi_schemas'
get_openapi_schemas_hash_uri = base_api_uri + '/get_openapi_schemas_hash'
get_openapi_url_uri = base_api_uri + '/get_openapi_url'
set_openapi_url_uri = base_api_uri + '/set_openapi_url'
get_openapi_url_hash_uri = base_api_uri + '/get_openapi_url_hash'
set_openapi_schema_paths_uri = base_api_uri + '/set_openapi_schema_paths'
get_openapi_schema_paths_uri = base_api_uri + '/get_openapi_schema_paths'
get_openapi_exclude_paths_uri = base_api_uri + '/get_openapi_exclude_paths'
set_openapi_exclude_paths_uri = base_api_uri + '/set_openapi_exclude_paths'
get_openapi_exclude_paths_hash_uri = base_api_uri + '/get_openapi_exclude_paths_hash'


def schema_name_after_change(old_name, new_name, waf_id):

    # update hash for schemas
    r = set_openapi_hash('openapi_schemas', waf_id)
    if r is False:
        return False

    # init database connection
    connect = init_db_connection()
    cursor = connect.cursor(cursor_factory=RealDictCursor)

    ##
    # Update the URL
    ##

    try:

        # search the data associated with the old schema name
        query = """ SELECT convert_from(url,'UTF8') AS url FROM openapi_url WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()

        # data processing
        if rows:

            try:
                url = json.loads(rows.get('url'))
            except json.JSONDecodeError:
                url = rows.strip('[]').split(',')

            url = [i.replace(old_name, new_name, 1) for i in url]

            try:
                url = json.dumps(url)
            except json.JSONDecodeError:
                url = str(url)

            query = """
                UPDATE
                    openapi_url
                SET
                    url = %s,
                    timestamp = CURRENT_TIMESTAMP
                WHERE
                    (waf_id = %s)
            """
            cursor.execute(query, (url, waf_id,))

            r = set_openapi_hash('openapi_url', waf_id)
            if r is False:
                close_db_connection(connect, cursor)
                return False

    except Exception as e:
        log.error('An error occurred while updating OpenAPI URL (WAF ID: {}): {}'.format(waf_id, e))

    ##
    # Update the paths
    ##

    try:

        # delete the data associated with the new schema name to avoid update conflicts.
        query = """ DELETE FROM openapi_schema_paths WHERE (waf_id = %s AND schema_name = %s) """
        cursor.execute(query, (waf_id, new_name,))

        # search the data associated with the old schema name
        query = """ SELECT id, url, schema_name FROM openapi_schema_paths WHERE (waf_id = %s AND schema_name = %s) """
        cursor.execute(query, (waf_id, old_name,))
        rows = cursor.fetchall()

        # data processing
        for i in rows:
            url = i.get('url', '').replace(old_name, new_name, 1)
            query = """
                UPDATE
                    openapi_schema_paths
                SET
                    url = %s,
                    schema_name = %s,
                    timestamp = CURRENT_TIMESTAMP
                WHERE
                    (id = %s)
            """
            cursor.execute(query, (url, new_name, i.get('id'),))

    except Exception as e:
        log.error('An error occurred while updating OpenAPI paths for schema {} (WAF ID: {}): {}'.format(old_name, waf_id, e))

    ##
    # Update the exclude paths
    ##

    try:

        # search the data associated with the old schema name
        query = """ SELECT convert_from(exclude_paths,'UTF8') AS exclude_paths FROM openapi_exclude_paths WHERE (waf_id = %s) """
        cursor.execute(query, (waf_id,))
        rows = cursor.fetchone()

        # data processing
        if rows:

            try:
                exclude_paths = json.loads(rows.get('exclude_paths'))
            except json.JSONDecodeError:
                exclude_paths = rows.strip('[]').split(',')

            exclude_paths = [i.replace(old_name, new_name, 1) for i in exclude_paths]

            try:
                exclude_paths = json.dumps(exclude_paths)
            except json.JSONDecodeError:
                exclude_paths = str(exclude_paths)

            query = """
                UPDATE
                    openapi_exclude_paths
                SET
                    exclude_paths = %s,
                    timestamp = CURRENT_TIMESTAMP
                WHERE
                    (waf_id = %s)
            """
            cursor.execute(query, (exclude_paths, waf_id,))

    except Exception as e:
        log.error(
            'An error occurred while updating OpenAPI exclude paths (WAF ID: {}): {}'
            .format(waf_id, e)
        )

    # close database connection
    close_db_connection(connect, cursor)

    # response
    r = set_openapi_hash('openapi_exclude_paths', waf_id)
    if r is False:
        return False
    else:
        return True


def set_openapi_hash(table, waf_id):

    # init database connection
    connect = init_db_connection()
    cursor = connect.cursor()

    if table == 'openapi_schemas':

        field = 'schema'
        memc_key = 'hash_schemas_' + waf_id
        query = """
            SELECT
                timestamp, convert_from(schema,'UTF8') AS schema
            FROM
                openapi_schemas
            WHERE
                waf_id = %s
            ORDER BY
                timestamp
            DESC
        """

    elif table == 'openapi_url':

        memc_key = 'hash_url_' + waf_id
        field = 'url'
        query = """
            SELECT
                timestamp, convert_from(url,'UTF8') AS url
            FROM
                openapi_url
            WHERE
                waf_id = %s
        """

    elif table == 'openapi_exclude_paths':

        memc_key = 'hash_exclude_path_' + waf_id
        field = 'exclude_paths'
        query = """
            SELECT
                timestamp, convert_from(exclude_paths,'UTF8') AS exclude_paths
            FROM
                openapi_exclude_paths
            WHERE
                waf_id = %s
        """

    else:
        log.error('An error occurred while generate hash: table {} is not valid'.format(table))
        close_db_connection(connect, cursor)
        return False

    cursor.execute(query, (waf_id,))
    rows = cursor.fetchall()

    h = waf_id
    if field == 'schema':
        if rows:
            timestamp = rows[0][0]
            h += str(timestamp)

        for i in rows:
            h += str(i[1])
    else:
        if rows:
            timestamp = rows[0][0]
            h += str(timestamp)
            try:
                rows = json.loads(rows[0][1])
            except json.JSONDecodeError:
                rows = rows[0][1].strip('[]').split(',')

        for i in rows:
            h += str(i)

    if rows:
        hash_elms = hashlib.sha256(h.encode()).hexdigest()
        with memc_lock:
            memc.set(memc_key, hash_elms)
    else:
        with memc_lock:
            memc.delete(memc_key)

    # close database connection
    close_db_connection(connect, cursor)

    # response
    return True


def oa_validation_rwo(data, status, status_fail):
    try:

        ##
        # OpenAPI readOnly/writeOnly simultaneous use validation
        ##

        # dict processing
        if isinstance(data, (dict, jsonref.JsonRef)):

            # if readOnly and writeOnly in data
            if 'readOnly' in data and 'writeOnly' in data:

                # if they are both True
                if data['readOnly'] and data['writeOnly']:
                    status = status_fail
                    return data, status, status_fail

            # processing the data values
            for k, v in data.copy().items():

                # dict processing
                if isinstance(v, (dict, jsonref.JsonRef)):
                    data[k], status, status_fail = oa_validation_rwo(v, status, status_fail)

                # list processing
                elif isinstance(v, list):

                    # reset the value
                    data[k] = []

                    # update the value and the status
                    for x in v:
                        if not isinstance(x, str):
                            r = oa_validation_rwo(v, status, status_fail)
                            data[k].append(r[0])
                            status = status_fail if r[1] == status_fail else status

                # other type processing
                else:
                    data[k] = v

        # finish processing
        else:
            return data, status, status_fail

    except Exception as e:
        log.error('An error occurred while OpenAPI schema validation (readOnly/writeOnly): {}'.format(e))

    return data, status, status_fail
